from __future__ import annotations

import base64
import json
from typing import Any

import boto3
from botocore.exceptions import ClientError

from x8.core import Context, DataModel, NCall, Operation, Provider, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingOperation,
    MessagingOperationParser,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)

from .._feature import PubSubFeature


class AmazonSNSBase(Provider):
    topic: str | None
    subscription: str | None
    region: str | None
    profile_name: str | None
    account_id: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    create_drop_queue: bool
    nparams: dict[str, Any]

    _session: Any
    _sns_client: Any
    _sqs_client: Any
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _account_id: str | None
    _queue_url_cache: dict[str, str]

    def __init__(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        region: str | None = None,
        account_id: str | None = None,
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        create_drop_queue: bool = True,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            topic:
                SNS topic name or ARN.
            subscription:
                SQS subscription name or ARN.
            region_name:
                AWS region name.
            account_id:
                AWS account ID.
            profile_name:
                AWS profile name.
            aws_access_key_id:
                AWS access key ID.
            aws_secret_access_key:
                AWS secret access key.
            aws_session_token:
                AWS session token.
            create_drop_queue:
                Whether to create and drop the queue
                on creating and dropping subscriptions.
                Defaults to True.
            nparams:
                Native parameters to AWS clients.
        """
        self.topic = topic
        self.subscription = subscription
        self.region = region
        self.account_id = account_id
        self.profile_name = profile_name
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_session_token = aws_session_token
        self.create_drop_queue = create_drop_queue
        self.nparams = nparams

        self._sns_client = None
        self._sqs_client = None
        self._account_id = None
        self._queue_url_cache = dict()

        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()
        super().__init__(**kwargs)

    def __supports__(self, feature: str) -> bool:
        return feature not in [PubSubFeature.BUILTIN_DLQ]

    def __setup__(self, context: Context | None = None) -> None:
        if self._sns_client is not None:
            return

        self._session = boto3.Session(
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            profile_name=self.profile_name,
            region_name=self.region,
        )

        self._sns_client = self._session.client("sns", **self.nparams)
        self._sqs_client = self._session.client("sqs", **self.nparams)

    def _get_topic_arn(self, op_parser: MessagingOperationParser) -> str:
        topic_name = self._get_topic_name(op_parser)
        if topic_name is None:
            raise BadRequestError("Topic name is required")

        if topic_name.startswith("arn:aws:sns:"):
            return topic_name

        account_id = self._get_account_id()
        return f"arn:aws:sns:{self.region}:{account_id}:{topic_name}"

    def _get_subscription_endpoint(
        self, op_parser: MessagingOperationParser
    ) -> str:
        subscription_name = self._get_subscription_name(op_parser)
        if subscription_name is None:
            raise BadRequestError("Subscription name is required")

        if subscription_name.startswith("arn:aws:"):
            return subscription_name

        account_id = self._get_account_id()
        return f"arn:aws:sqs:{self.region}:{account_id}:{subscription_name}"

    def _get_account_id(self) -> str:
        if self._account_id is not None:
            return self._account_id
        sts_client = self._session.client("sts", region_name=self.region)
        response = sts_client.get_caller_identity()
        self._account_id = response["Account"]
        return self._account_id

    def _get_topic_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        topic_name = op_parser.get_topic()
        topic_name = topic_name or self.topic or self.__component__.topic
        return topic_name

    def _get_subscription_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        subscription_name = op_parser.get_subscription()
        subscription_name = (
            subscription_name
            or self.subscription
            or self.__component__.subscription
        )
        return subscription_name

    def _get_queue_url(self, op_parser: MessagingOperationParser) -> str:
        subscription_name = self._get_subscription_name(op_parser)
        if subscription_name is None:
            raise BadRequestError("Subscription name is required")

        if subscription_name.startswith("arn:aws:sqs:"):
            queue_name = subscription_name.split(":")[-1]
        else:
            queue_name = subscription_name

        if queue_name in self._queue_url_cache:
            return self._queue_url_cache[queue_name]

        response = self._sqs_client.get_queue_url(QueueName=queue_name)
        queue_url = response["QueueUrl"]
        self._queue_url_cache[queue_name] = queue_url
        return queue_url

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        ncall = self._get_ncall(
            op_parser,
            ClientHelper(
                self._sns_client,
                self._sqs_client,
                self._op_converter,
            ),
        )
        if ncall is None:
            return super().__run__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        client_helper: ClientHelper,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()
        op_converter = self._op_converter

        # CREATE TOPIC
        if op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            args: dict = {
                "topic_name": self._get_topic_name(op_parser),
                "config": op_parser.get_topic_config(),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.create_topic,
                args,
                None,
            )
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            args = {
                "topic_arn": self._get_topic_name(op_parser),
                "config": op_parser.get_topic_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.update_topic,
                args,
                None,
            )
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.drop_topic,
                args,
            )
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.has_topic,
                args,
                None,
            )
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.get_topic,
                args,
                None,
            )
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            call = NCall(self._sns_client.list_topics, {}, nargs)
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "protocol": "sqs",
                "endpoint": self._get_subscription_endpoint(op_parser),
                "config": op_parser.get_subscription_config(),
                "where_exists": op_parser.get_where_exists(),
                "create_drop_queue": self.create_drop_queue,
                "nargs": nargs,
            }
            call = NCall(
                client_helper.create_subscription,
                args,
                None,
            )
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "endpoint": self._get_subscription_endpoint(op_parser),
                "queue_url": self._get_queue_url(op_parser),
                "config": op_parser.get_subscription_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.update_subscription,
                args,
                None,
            )
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "endpoint": self._get_subscription_endpoint(op_parser),
                "queue_url": self._get_queue_url(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "create_drop_queue": self.create_drop_queue,
                "nargs": nargs,
            }
            call = NCall(
                client_helper.drop_subscription,
                args,
                None,
            )
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            args = {
                "TopicArn": self._get_topic_arn(op_parser),
            }
            call = NCall(
                self._sns_client.list_subscriptions_by_topic,
                args,
                None,
            )
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "endpoint": self._get_subscription_endpoint(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.has_subscription,
                args,
                None,
            )
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            args = {
                "topic_arn": self._get_topic_arn(op_parser),
                "endpoint": self._get_subscription_endpoint(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.get_subscription,
                args,
                None,
            )
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = op_converter.convert_put(
                self._get_topic_arn(op_parser),
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            call = NCall(
                self._sns_client.publish,
                args,
                nargs,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = {
                "batch": op_parser.get_batch(),
                "topic_arn": self._get_topic_arn(op_parser),
            }
            call = NCall(
                client_helper.batch,
                args,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            args = op_converter.convert_pull(
                self._get_queue_url(op_parser),
                op_parser.get_pull_config(),
            )
            call = NCall(
                self._sqs_client.receive_message,
                args,
                nargs,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = {
                "QueueUrl": self._get_queue_url(op_parser),
                "ReceiptHandle": op_parser.get_key().nref,
            }
            call = NCall(
                self._sqs_client.delete_message,
                args,
                nargs,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = {
                "QueueUrl": self._get_queue_url(op_parser),
                "ReceiptHandle": op_parser.get_key().nref,
                "VisibilityTimeout": 0,
            }
            call = NCall(
                self._sqs_client.change_message_visibility,
                args,
                nargs,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = {
                "QueueUrl": self._get_queue_url(op_parser),
                "ReceiptHandle": op_parser.get_key().nref,
                "VisibilityTimeout": op_parser.get_timeout(),
            }
            call = NCall(
                self._sqs_client.change_message_visibility,
                args,
                nargs,
            )
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            args = {
                "queue_url": self._get_queue_url(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.purge,
                args,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {
                "topic": self._get_topic_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(client_helper.close, args)
        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result_converter = self._result_converter
        result: Any = None
        # CREATE TOPIC
        if op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            result = None
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            result = None
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            result = None
        # LIST TOPICS
        if op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            result = result_converter.convert_list_topics(nresult)
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            result = nresult
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            result = nresult
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            result = None
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            result = None
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            result = None
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            result = result_converter.convert_list_subscriptions(nresult)
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            result = nresult
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            result = nresult
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = result_converter.convert_pull(nresult)
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            result = MessageItem(key=op_parser.get_key())
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            result = None
        return result


class ClientHelper:
    sns_client: Any
    sqs_client: Any
    op_converter: OperationConverter

    def __init__(
        self,
        sns_client: Any,
        sqs_client: Any,
        op_converter: OperationConverter,
    ):
        self.sns_client = sns_client
        self.sqs_client = sqs_client
        self.op_converter = op_converter

    def create_topic(
        self,
        topic_name: str,
        config: TopicConfig | None,
        where_exists: bool | None,
        nargs: Any,
    ) -> Any:
        if where_exists is False:
            response = self.sns_client.list_topics()
            for topic in response["Topics"]:
                if topic["TopicArn"].split(":")[-1] == topic_name:
                    raise ConflictError

        args = self.op_converter.convert_create_topic(
            topic_name=topic_name, config=config
        )
        NCall(
            self.sns_client.create_topic,
            args,
            nargs,
        ).invoke()

    def update_topic(
        self,
        topic_arn: str,
        config: TopicConfig | None,
        nargs: Any,
    ) -> None:
        if config is None:
            return
        args = {
            "TopicArn": topic_arn,
        }
        if config.nconfig:
            for key, value in config.nconfig.items():
                args["AttributeName"] = key
                args["AttributeValue"] = value
                NCall(
                    self.sns_client.set_topic_attributes,
                    args,
                    nargs,
                ).invoke()

    def drop_topic(
        self,
        topic_arn: str,
        where_exists: bool | None,
        nargs: Any,
    ) -> None:
        if where_exists is True:
            try:
                self.sns_client.get_topic_attributes(TopicArn=topic_arn)
            except ClientError as e:
                if e.response["Error"]["Code"] == "NotFound":
                    raise NotFoundError
                else:
                    raise
        NCall(
            self.sns_client.delete_topic,
            {"TopicArn": topic_arn},
            nargs,
        ).invoke()

    def has_topic(
        self,
        topic_arn: str,
        nargs: Any,
    ) -> bool:
        try:
            self.sns_client.get_topic_attributes(TopicArn=topic_arn)
        except ClientError as e:
            if e.response["Error"]["Code"] == "NotFound":
                return False
            else:
                raise
        return True

    def get_topic(
        self,
        topic_arn: str,
        nargs: Any,
    ) -> TopicInfo:
        try:
            response = NCall(
                self.sns_client.get_topic_attributes,
                {"TopicArn": topic_arn},
                nargs,
            ).invoke()
        except ClientError as e:
            if e.response["Error"]["Code"] == "NotFound":
                raise NotFoundError
            else:
                raise
        return TopicInfo(
            name=topic_arn.split(":")[-1],
            subscription_count=int(
                response["Attributes"].get("SubscriptionsConfirmed", 0)
            ),
            nref=topic_arn,
        )

    def create_subscription(
        self,
        topic_arn: str,
        protocol: str,
        endpoint: str,
        config: SubscriptionConfig | None,
        where_exists: bool | None,
        create_drop_queue: bool,
        nargs: Any,
    ) -> Any:
        if protocol == "sqs" and endpoint.startswith("arn:aws:sqs:"):
            queue_name = endpoint.split(":")[-1]
            try:
                self.sqs_client.get_queue_url(QueueName=queue_name)
            except ClientError as e:
                if (
                    e.response["Error"]["Code"]
                    == "AWS.SimpleQueueService.NonExistentQueue"
                ):
                    if not create_drop_queue:
                        raise
                    queue_attributes = (
                        self.op_converter.convert_queue_attributes(
                            topic_arn=topic_arn,
                            endpoint=endpoint,
                            config=config,
                        )
                    )
                    self.sqs_client.create_queue(
                        QueueName=queue_name, Attributes=queue_attributes
                    )
        if where_exists is False:
            response = self.sns_client.list_subscriptions_by_topic(
                TopicArn=topic_arn
            )
            for subscription in response["Subscriptions"]:
                if subscription["Endpoint"] == endpoint:
                    raise ConflictError

        args = self.op_converter.convert_create_subscription(
            topic_arn=topic_arn,
            protocol=protocol,
            endpoint=endpoint,
            config=config,
        )
        return NCall(
            self.sns_client.subscribe,
            args,
            nargs,
        ).invoke()

    def update_subscription(
        self,
        topic_arn: str,
        endpoint: str,
        queue_url: str,
        config: SubscriptionConfig | None,
        nargs: Any,
    ) -> None:
        attributes = self.op_converter.convert_queue_attributes(
            topic_arn=topic_arn, endpoint=endpoint, config=config
        )
        NCall(
            self.sqs_client.set_queue_attributes,
            {
                "QueueUrl": queue_url,
                "Attributes": attributes,
            },
            nargs,
        ).invoke()

    def drop_subscription(
        self,
        topic_arn: str,
        endpoint: str,
        queue_url: str,
        where_exists: bool | None,
        create_drop_queue: bool,
        nargs: Any,
    ) -> None:
        response = self.sns_client.list_subscriptions_by_topic(
            TopicArn=topic_arn
        )
        subscription_arn = None
        for subscription in response["Subscriptions"]:
            if subscription["Endpoint"] == endpoint:
                subscription_arn = subscription["SubscriptionArn"]
        if not subscription_arn:
            if where_exists is True:
                raise NotFoundError
            return

        if create_drop_queue:
            subscription_details = self.sns_client.get_subscription_attributes(
                SubscriptionArn=subscription_arn
            )
        self.sns_client.unsubscribe(SubscriptionArn=subscription_arn)
        if create_drop_queue:
            if (
                "Attributes" in subscription_details
                and "Protocol" in subscription_details["Attributes"]
                and subscription_details["Attributes"]["Protocol"] == "sqs"
                and "Endpoint" in subscription_details["Attributes"]
            ):
                endpoint = subscription_details["Attributes"]["Endpoint"]
                if endpoint.startswith("arn:aws:sqs:"):
                    self.sqs_client.delete_queue(QueueUrl=queue_url)

    def has_subscription(
        self,
        topic_arn: str,
        endpoint: str,
        nargs: Any,
    ) -> bool:
        response = self.sns_client.list_subscriptions_by_topic(
            TopicArn=topic_arn
        )
        for subscription in response["Subscriptions"]:
            if subscription["Endpoint"] == endpoint:
                return True
        return False

    def get_subscription(
        self,
        topic_arn: str,
        endpoint: str,
        nargs: Any,
    ) -> SubscriptionInfo:
        try:
            queue_name = endpoint.split(":")[-1]
            res = self.sqs_client.get_queue_url(QueueName=queue_name)
            queue_url = res["QueueUrl"]
            res = NCall(
                self.sqs_client.get_queue_attributes,
                {
                    "QueueUrl": queue_url,
                    "AttributeNames": [
                        "All",
                    ],
                },
                nargs,
            ).invoke()
            attributes = res["Attributes"]
            fifo = attributes.get("FifoQueue", "false").lower() == "true"
            queue_arn = attributes.get("QueueArn")
            max_receive_count = None
            dlq_nref = None
            if "RedrivePolicy" in attributes:
                redrive_policy = json.loads(attributes["RedrivePolicy"])
                dlq_nref = redrive_policy.get("deadLetterTargetArn")
                if "maxReceiveCount" in redrive_policy:
                    max_receive_count = int(
                        redrive_policy.get("maxReceiveCount", 0)
                    )
            return SubscriptionInfo(
                name=endpoint.split(":")[-1],
                topic=topic_arn.split(":")[-1],
                active_message_count=int(
                    attributes.get("ApproximateNumberOfMessages", 0)
                ),
                inflight_message_count=int(
                    attributes.get("ApproximateNumberOfMessagesNotVisible", 0)
                ),
                scheduled_message_count=int(
                    attributes.get("ApproximateNumberOfMessagesDelayed", 0)
                ),
                config=SubscriptionConfig(
                    visibility_timeout=int(
                        attributes.get("VisibilityTimeout", 0)
                    ),
                    ttl=int(attributes.get("MessageRetentionPeriod", 0)),
                    max_delivery_count=max_receive_count,
                    dlq_nref=dlq_nref,
                    fifo=fifo,
                ),
                nref=queue_arn,
            )
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                raise NotFoundError
            raise

    def batch(
        self,
        batch: MessageBatch,
        topic_arn: str,
    ) -> list:
        entries = []
        for i, operation in enumerate(batch.operations):
            op_parser = MessagingOperationParser(operation)
            args = self.op_converter.convert_put(
                topic_arn,
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            entry = {
                "Id": str(i),
                "Message": args["Message"],
            }

            if "MessageAttributes" in args:
                entry["MessageAttributes"] = args["MessageAttributes"]

            if "MessageGroupId" in args:
                entry["MessageGroupId"] = args["MessageGroupId"]

            if "MessageDeduplicationId" in args:
                entry["MessageDeduplicationId"] = args[
                    "MessageDeduplicationId"
                ]

            entries.append(entry)
        response = NCall(
            self.sns_client.publish_batch,
            {"TopicArn": topic_arn, "PublishBatchRequestEntries": entries},
        ).invoke()

        message_ids = []
        for msg in response.get("Successful", []):
            message_ids.append(msg.get("MessageId"))

        if "Failed" in response and response["Failed"]:
            for failed in response["Failed"]:
                raise BadRequestError(
                    (
                        f"Failed to send message {failed['Id']}: "
                        f"{failed['Message']}"
                    )
                )
        return message_ids

    def purge(
        self,
        queue_url: str,
        nargs: Any,
    ) -> None:
        try:
            NCall(
                self.sqs_client.purge_queue,
                {"QueueUrl": queue_url},
                nargs,
            ).invoke()
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.PurgeQueueInProgress"
            ):
                pass

    def close(self, topic: str | None, nargs: Any) -> Any:
        return None


class OperationConverter:
    def convert_create_topic(
        self,
        topic_name: str,
        config: TopicConfig | None,
    ) -> dict:
        args: dict = {
            "Name": topic_name,
        }
        if config and config.nconfig:
            args["Attributes"] = config.nconfig
        return args

    def convert_queue_attributes(
        self,
        topic_arn: str,
        endpoint: str,
        config: SubscriptionConfig | None,
    ) -> dict:
        attributes = {}
        attributes["Policy"] = json.dumps(
            {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "sns.amazonaws.com"},
                        "Action": "sqs:SendMessage",
                        "Resource": endpoint,
                        "Condition": {
                            "ArnEquals": {"aws:SourceArn": topic_arn}
                        },
                    }
                ],
            }
        )
        if config:
            if config.visibility_timeout:
                attributes["VisibilityTimeout"] = str(
                    int(config.visibility_timeout)
                )
            if config.ttl:
                attributes["MessageRetentionPeriod"] = str(int(config.ttl))
            if config.fifo:
                attributes["FifoQueue"] = "true"
            if config.dlq_nref and config.max_delivery_count:
                redrive_policy = {
                    "deadLetterTargetArn": config.dlq_nref,
                    "maxReceiveCount": str(config.max_delivery_count),
                }
                attributes["RedrivePolicy"] = json.dumps(redrive_policy)
            if config.nconfig:
                attributes.update(config.nconfig)
        return attributes

    def convert_create_subscription(
        self,
        topic_arn: str,
        protocol: str,
        endpoint: str,
        config: SubscriptionConfig | None,
    ) -> dict:
        args = {
            "TopicArn": topic_arn,
            "Protocol": protocol,
            "Endpoint": endpoint,
        }
        return args

    def convert_put(
        self,
        topic_arn: str,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
        config: MessagePutConfig | None,
    ) -> dict:
        if config is not None and config.delay is not None:
            raise BadRequestError("Delay is not supported for SNS publish")

        message_body: str = ""
        message_attributes: dict = {}
        if isinstance(value, str):
            message_body = value
            content_type = "text/plain"
        elif isinstance(value, bytes):
            message_body = base64.b64encode(value).decode("utf-8")
            content_type = "application/octet-stream"
        elif isinstance(value, dict):
            message_body = json.dumps(value)
            content_type = "application/json"
        elif isinstance(value, DataModel):
            message_body = value.to_json()
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")

        message_attributes["content_type"] = {
            "DataType": "String",
            "StringValue": content_type,
        }

        if metadata:
            for key, val in metadata.items():
                if isinstance(val, str):
                    message_attributes[key] = {
                        "DataType": "String",
                        "StringValue": val,
                    }
                elif isinstance(val, (int, float)):
                    message_attributes[key] = {
                        "DataType": "Number",
                        "StringValue": str(val),
                    }
                elif isinstance(val, bytes):
                    message_attributes[key] = {
                        "DataType": "Binary",
                        "BinaryValue": val,
                    }
                else:
                    message_attributes[key] = {
                        "DataType": "String",
                        "StringValue": str(val),
                    }

        message_deduplication_id = None
        message_group_id = None
        if properties:
            if properties.message_id:
                message_attributes["message_id"] = {
                    "DataType": "String",
                    "StringValue": properties.message_id,
                }
                message_deduplication_id = properties.message_id

            if properties.group_id:
                message_attributes["group_id"] = {
                    "DataType": "String",
                    "StringValue": properties.group_id,
                }
                message_group_id = properties.group_id

        args: dict = {
            "TopicArn": topic_arn,
            "Message": message_body,
        }

        if message_attributes:
            args["MessageAttributes"] = message_attributes

        if message_group_id and topic_arn.endswith(".fifo"):
            args["MessageGroupId"] = message_group_id

        if message_deduplication_id and topic_arn.endswith(".fifo"):
            args["MessageDeduplicationId"] = message_deduplication_id

        return args

    def convert_pull(
        self, queue_url: str, config: MessagePullConfig | None
    ) -> dict:
        max_messages = 1
        if config is not None and config.max_count is not None:
            max_messages = min(config.max_count, 10)  # SQS allows max 10

        args = {
            "QueueUrl": queue_url,
            "MaxNumberOfMessages": max_messages,
            "MessageAttributeNames": ["All"],
            "AttributeNames": ["All"],
        }

        if config is not None and config.max_wait_time is not None:
            args["WaitTimeSeconds"] = min(
                int(config.max_wait_time), 20
            )  # SQS allows max 20
        if config is not None and config.visibility_timeout is not None:
            args["VisibilityTimeout"] = int(config.visibility_timeout)

        return args


class ResultConverter:
    def _convert_key(self, message: dict) -> MessageKey:
        return MessageKey(nref=message.get("ReceiptHandle"))

    def _convert_value(self, message: dict) -> MessageValueType:
        message_attributes = message.get("MessageAttributes", {})
        content_type = message_attributes.get(
            "content_type", {"Value": "text/plain"}
        ).get("Value", "text/plain")

        body = message.get("Message", "")

        if content_type == "application/octet-stream":
            return base64.b64decode(body)
        elif content_type == "application/json":
            try:
                return json.loads(body)
            except json.JSONDecodeError:
                return body

        return body

    def _convert_metadata(self, message: dict) -> dict | None:
        message_attributes = message.get("MessageAttributes", {})
        if not message_attributes:
            return None

        metadata = {}
        for key, attr in message_attributes.items():
            # Skip internal attributes
            if key in ["content_type", "message_id", "group_id"]:
                continue

            if attr.get("Type") == "String":
                metadata[key] = attr.get("Value")
            elif attr.get("Type") == "Number":
                try:
                    metadata[key] = float(attr.get("Value"))
                    if metadata[key].is_integer():
                        metadata[key] = int(metadata[key])
                except (ValueError, TypeError):
                    metadata[key] = attr.get("Value")
            elif attr.get("Type") == "Binary":
                metadata[key] = attr.get("Value")
            else:
                metadata[key] = attr.get("Value")

        return metadata if metadata else None

    def _convert_properties(
        self, message: dict, attributes: dict
    ) -> MessageProperties:
        message_attributes = message.get("MessageAttributes", {})
        message_id = None
        if "message_id" in message_attributes:
            message_id = message_attributes["message_id"].get("Value")

        group_id = None
        if "group_id" in message_attributes:
            group_id = message_attributes["group_id"].get("Value")

        content_type = "text/plain"
        if "content_type" in message_attributes:
            content_type = message_attributes["content_type"].get(
                "Value", "text/plain"
            )

        enqueued_time = None
        if "SentTimestamp" in attributes:
            try:
                enqueued_time = float(attributes["SentTimestamp"]) / 1000.0
            except (ValueError, TypeError):
                pass

        # Get other properties
        return MessageProperties(
            message_id=message_id or message.get("MessageId"),
            content_type=content_type,
            group_id=group_id,
            enqueued_time=enqueued_time,
            delivery_count=(int(attributes.get("ApproximateReceiveCount", 1))),
        )

    def _convert_message(
        self,
        message: dict,
    ) -> MessageItem:
        message_body = json.loads(message.get("Body", "{}"))
        attributes = message.get("Attributes", {})
        return MessageItem(
            key=self._convert_key(message=message),
            value=self._convert_value(message=message_body),
            metadata=self._convert_metadata(message=message_body),
            properties=self._convert_properties(
                message=message_body, attributes=attributes
            ),
        )

    def convert_list_topics(self, nresult: Any) -> list[str]:
        topics = []
        for topic in nresult.get("Topics", []):
            topic_name = topic["TopicArn"].split(":")[-1]
            topics.append(topic_name)
        return topics

    def convert_list_subscriptions(self, nresult: Any) -> list[str]:
        subscriptions = []
        for subscription in nresult.get("Subscriptions", []):
            subscriptions.append(subscription["Endpoint"].split(":")[-1])
        return subscriptions

    def convert_pull(self, nresult: Any) -> list[MessageItem]:
        result = []
        for message in nresult.get("Messages", []):
            result.append(self._convert_message(message))
        return result



import ast
import textwrap
from typing import Any

from x8.core import Context, Operation, Provider


class Default(Provider):
    script: str
    _kwargs: Any

    def __init__(self, script: str, **kwargs):
        self.script = script
        self._kwargs = kwargs

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        # Parse and reformat the script safely
        try:
            ast.parse(self.script)
        except SyntaxError as e:
            raise ValueError(f"Invalid script: {e}")

        execution_context = {**self._kwargs}
        script_body = textwrap.indent(self.script, "    ")
        script_function = f"""
def __run__(operation, context):
{script_body}
"""
        exec(script_function, execution_context)
        return execution_context["__run__"](operation, context)

    async def __arun__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        # Parse and reformat the script safely
        try:
            ast.parse(self.script)
        except SyntaxError as e:
            raise ValueError(f"Invalid script: {e}")

        execution_context = {**self._kwargs}
        script_body = textwrap.indent(self.script, "    ")
        script_function = f"""
async def __run__(operation, context):
{script_body}
"""
        exec(script_function, execution_context)
        return await execution_context["__run__"](operation, context)



from typing import Any

from x8.messaging._common import MessagingMode
from x8.messaging._common.redis import RedisBase

from .._feature import PubSubFeature


class Redis(RedisBase):
    def __init__(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        url: str | None = None,
        host: str | None = None,
        port: int | None = None,
        db: int = 0,
        username: str | None = None,
        password: str | None = None,
        options: dict | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
            url:
                Redis URL.
            host:
                Redis host.
            port:
                Redis port.
            db:
                Redis database. Defaults to 0.
            username:
                Redis username.
            password:
                Redis password.
            options:
                Native parameters to Redis client.
        """
        super().__init__(
            mode=MessagingMode.PUBSUB,
            topic=topic,
            subscription=subscription,
            url=url,
            host=host,
            port=port,
            db=db,
            username=username,
            password=password,
            options=options,
            nparams=nparams,
            **kwargs,
        )

    def __supports__(self, feature: str) -> bool:
        return feature not in [
            PubSubFeature.BUILTIN_DLQ,
            PubSubFeature.CONFIGURABLE_DLQ,
        ]



from typing import Any

from common.secrets import get_secrets

from x8.compute.kubernetes import Kubernetes

secrets = get_secrets()


class KubernetesProvider:
    AZURE_KUBERNETES_SERVICE = "azure_kubernetes_service"
    MINIKUBE = "minikube"


provider_parameters: dict[str, dict[str, Any]] = {
    KubernetesProvider.AZURE_KUBERNETES_SERVICE: {},
    KubernetesProvider.MINIKUBE: {},
}


def get_component(provider_type: str):
    component = Kubernetes(
        __provider__=dict(
            type=provider_type,
            parameters=provider_parameters[provider_type],
        ),
    )
    return component



from typing import Any

from x8.core import Component, Response, operation
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)
from x8.ql import Expression


class PubSub(Component):
    topic: str | None
    subscription: str | None

    def __init__(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        self.topic = topic
        self.subscription = subscription
        super().__init__(**kwargs)

    @operation()
    def create_topic(
        self,
        topic: str | None = None,
        config: dict | TopicConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create topic.

        Args:
            topic:
                Topic name.
            config:
                Topic config.
            where:
                Condition expression.
        """
        raise NotImplementedError

    @operation()
    def update_topic(
        self,
        topic: str | None = None,
        config: dict | TopicConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Update topic.

        Args:
            topic:
                Topic name.
            config:
                Topic config.
        """
        raise NotImplementedError

    @operation()
    def drop_topic(
        self,
        topic: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop topic.

        Args:
            topic:
                Topic name.
            where:
                Condition expression.
        """
        raise NotImplementedError

    @operation()
    def list_topics(self, **kwargs: Any) -> Response[list[str]]:
        """List topics.

        Returns:
            List of topic names.
        """
        raise NotImplementedError

    @operation()
    def has_topic(
        self,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if topic exists.

        Args:
            topic:
                Topic name.

        Returns:
            True if topic exists, False otherwise.
        """
        raise NotImplementedError

    @operation()
    def get_topic(
        self, topic: str | None = None, **kwargs: Any
    ) -> Response[TopicInfo]:
        """Get topic.

        Args:
            topic:
                Topic name.

        Returns:
            Topic info.
        """
        raise NotImplementedError

    @operation()
    def create_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        config: dict | SubscriptionConfig | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
            config:
                Subscription config.
            where:
                Condition expression.
        """
        raise NotImplementedError

    @operation()
    def update_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        config: dict | SubscriptionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Update subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
            config:
                Subscription config.
        """
        raise NotImplementedError

    @operation()
    def drop_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        where: str | Expression | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
            where:
                Condition expression.
        """
        raise NotImplementedError

    @operation()
    def list_subscriptions(
        self,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List subscriptions.

        Returns:
            List of subscription names.
        """
        raise NotImplementedError

    @operation()
    def has_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if subscription exists.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            True if subscription exists, False otherwise.
        """
        raise NotImplementedError

    @operation()
    def get_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        **kwargs: Any,
    ) -> Response[SubscriptionInfo]:
        """Get subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            Subscription info.
        """
        raise NotImplementedError

    @operation()
    def put(
        self,
        value: MessageValueType,
        metadata: dict | None = None,
        properties: dict | MessageProperties | None = None,
        config: dict | MessagePutConfig | None = None,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[MessageItem]:
        """Put message in the topic.

        Args:
            value:
                Message value.
            metadata:
                Message metadata.
            properties:
                Message properties.
            config:
                Put config.
            topic:
                Topic name.

        Returns:
            Message item.
        """
        raise NotImplementedError

    @operation()
    def batch(
        self,
        batch: dict | MessageBatch,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[list[MessageItem]]:
        """Put batch of message in the topic.

        Args:
            batch: Message batch.
            topic:
                Topic name.

        Returns:
            List of message items.
        """
        raise NotImplementedError

    @operation()
    def pull(
        self,
        config: dict | MessagePullConfig | None = None,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[list[MessageItem]]:
        """Pull messages from the topic and subscription.

        Args:
            config:
                Pull config. Defaults to None.
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            Pulled message items.
        """
        raise NotImplementedError

    @operation()
    def ack(
        self,
        key: dict | MessageKey,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[None]:
        """Acknowledge message.

        Args:
            key:
                Message key.
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    def nack(
        self,
        key: dict | MessageKey,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[None]:
        """Abandon message.

        Args:
            key:
                Message key.
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    def extend(
        self,
        key: dict | MessageKey,
        timeout: int | None = None,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[MessageItem]:
        """Extend message timeout.

        Args:
            key:
                Message key.
            timeout:
                Number of seconds to extend the message timeout.
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            Message item.
        """
        raise NotImplementedError

    @operation()
    def purge(
        self,
        config: dict | MessagePullConfig | None = None,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[None]:
        """Purge all messages in the queue.

        Args:
            config:
                Pull config.
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    def close(
        self,
        topic: str | None = None,
    ) -> Response[None]:
        """Close the topic and the client.

        Args:
            topic:
                Topic name.
        """
        raise NotImplementedError

    @operation()
    async def acreate_topic(
        self,
        topic: str | None = None,
        config: dict | TopicConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create topic.

        Args:
            topic:
                Topic name.
            config:
                Topic config.
        """
        raise NotImplementedError

    @operation()
    async def aupdate_topic(
        self,
        topic: str | None = None,
        config: dict | TopicConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Update topic.

        Args:
            topic:
                Topic name.
            config:
                Topic config.
        """
        raise NotImplementedError

    @operation()
    async def adrop_topic(
        self, topic: str | None = None, **kwargs: Any
    ) -> Response[None]:
        """Drop topic.

        Args:
            topic:
                Topic name.
        """
        raise NotImplementedError

    @operation()
    async def alist_topics(self, **kwargs: Any) -> Response[list[str]]:
        """List topics.

        Returns:
            List of topic names.
        """
        raise NotImplementedError

    @operation()
    async def ahas_topic(
        self,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if topic exists.

        Args:
            topic:
                Topic name.

        Returns:
            True if topic exists, False otherwise.
        """
        raise NotImplementedError

    @operation()
    async def aget_topic(
        self, topic: str | None = None, **kwargs: Any
    ) -> Response[TopicInfo]:
        """Get topic.

        Args:
            topic:
                Topic name.

        Returns:
            Topic info.
        """
        raise NotImplementedError

    @operation()
    async def acreate_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        config: dict | SubscriptionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Create subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
            config:
                Subscription config.
        """
        raise NotImplementedError

    @operation()
    async def aupdate_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        config: dict | SubscriptionConfig | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Update subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
            config:
                Subscription config.
        """
        raise NotImplementedError

    @operation()
    async def adrop_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        **kwargs: Any,
    ) -> Response[None]:
        """Drop subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    async def alist_subscriptions(
        self,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[list[str]]:
        """List subscriptions.

        Returns:
            List of subscription names.
        """
        raise NotImplementedError

    @operation()
    async def ahas_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        **kwargs: Any,
    ) -> Response[bool]:
        """Check if subscription exists.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            True if subscription exists, False otherwise.
        """
        raise NotImplementedError

    @operation()
    async def aget_subscription(
        self,
        topic: str | None = None,
        subscription: str | None = None,
        **kwargs: Any,
    ) -> Response[SubscriptionInfo]:
        """Get subscription.

        Args:
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            Subscription info.
        """
        raise NotImplementedError

    @operation()
    async def aput(
        self,
        value: MessageValueType,
        metadata: dict | None = None,
        properties: dict | MessageProperties | None = None,
        config: dict | MessagePutConfig | None = None,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[MessageItem]:
        """Put message in the topic.

        Args:
            value:
                Message value.
            metadata:
                Message metadata.
            properties:
                Message properties.
            config:
                Put config.
            topic:
                Topic name.

        Returns:
            Message item.
        """
        raise NotImplementedError

    @operation()
    async def abatch(
        self,
        batch: dict | MessageBatch,
        topic: str | None = None,
        **kwargs: Any,
    ) -> Response[list[MessageItem]]:
        """Put batch of message in the topic.

        Args:
            batch: Message batch.
            topic:
                Topic name.

        Returns:
            List of message items.
        """
        raise NotImplementedError

    @operation()
    async def apull(
        self,
        config: dict | MessagePullConfig | None = None,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[list[MessageItem]]:
        """Pull messages from the topic and subscription.

        Args:
            config:
                Pull config. Defaults to None.
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            Pulled message items.
        """
        raise NotImplementedError

    @operation()
    async def aack(
        self,
        key: dict | MessageKey,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[None]:
        """Acknowledge message.

        Args:
            key:
                Message key.
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    async def anack(
        self,
        key: dict | MessageKey,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[None]:
        """Abandon message.

        Args:
            key:
                Message key.
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    async def aextend(
        self,
        key: dict | MessageKey,
        timeout: int | None = None,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[MessageItem]:
        """Extend message timeout.

        Args:
            key:
                Message key.
            timeout:
                Number of seconds to extend the message timeout.
            topic:
                Topic name.
            subscription:
                Subscription name.

        Returns:
            Message item.
        """
        raise NotImplementedError

    @operation()
    async def apurge(
        self,
        config: dict | MessagePullConfig | None = None,
        topic: str | None = None,
        subscription: str | None = None,
    ) -> Response[None]:
        """Purge all messages in the queue.

        Args:
            config:
                Pull config.
            topic:
                Topic name.
            subscription:
                Subscription name.
        """
        raise NotImplementedError

    @operation()
    async def aclose(
        self,
        topic: str | None = None,
    ) -> Response[None]:
        """Close the topic and the client.

        Args:
            topic:
                Topic name.
        """
        raise NotImplementedError



import os
import subprocess
from typing import Any, Literal

from x8._common.google_provider import GoogleProvider
from x8.compute.containerizer import DEFAULT_PLATFORM
from x8.core import Context, DataModel, Response
from x8.core.exceptions import BadRequestError, NotFoundError

from .._models import ContainerRegistryItem, ContainerRegistryItemDigest


class GoogleArtifactRegistryResource(DataModel):
    id: str
    name: str
    location: str
    login_server: str


class GoogleArtifactRegistry(GoogleProvider):
    project: str | None
    location: str | None
    name: str | None
    platform: str
    nparams: dict[str, Any] | None

    _client: Any
    _aclient: Any
    _credentials: Any
    _shell: bool = False
    _init: bool = False
    _ainit: bool = False

    def __init__(
        self,
        project: str | None = None,
        location: str = "us-central1",
        name: str | None = None,
        platform: str = DEFAULT_PLATFORM,
        service_account_info: str | None = None,
        service_account_file: str | None = None,
        access_token: str | None = None,
        nparams: dict[str, Any] | None = None,
        **kwargs,
    ):
        """Initialize.

        Args:
            project:
                Google Cloud project ID.
            location:
                Google Cloud location.
            name:
                Google Artifact Registry repository name.
            platform:
                Platform to use for the container when pulling.
            service_account_info:
                Google service account info with serialized credentials.
            service_account_file:
                Google service account file with credentials.
            access_token:
                Google access token.
            nparams:
                Native params to google client.
        """
        self.project = project
        self.location = location
        self.name = name
        self.platform = platform
        self.nparams = nparams
        self._init = False
        self._ainit = False
        self._shell = os.name == "nt"
        super().__init__(
            service_account_info=service_account_info,
            service_account_file=service_account_file,
            access_token=access_token,
            **kwargs,
        )

    def __setup__(self, context: Context | None = None) -> None:
        if self._init:
            return

        from google.cloud import artifactregistry_v1

        self._client = artifactregistry_v1.ArtifactRegistryClient(
            credentials=self._get_credentials(),
            **(self.nparams or {}),
        )
        self._init = True

    async def __asetup__(self, context: Context | None = None) -> None:
        if self._ainit:
            return

        from google.cloud import artifactregistry_v1

        self._aclient = artifactregistry_v1.ArtifactRegistryAsyncClient(
            credentials=self._get_credentials(),
            **(self.nparams or {}),
        )
        self._ainit = True

    def _artifact_registry_path(self) -> str:
        return f"{self.location}-docker.pkg.dev"

    def _repository_path(self) -> str:
        return (
            f"projects/{self._get_project_or_default(self.project)}/"
            f"locations/{self.location}/"
            f"repositories/{self.name}"
        )

    def _full_image_path(self, image_name: str, tag: str | None = None) -> str:
        if self.name:
            image = (
                f"{self._artifact_registry_path()}/"
                f"{self._get_project_or_default(self.project)}/"
                f"{self.name}/{image_name}"
            )
        else:
            image = (
                f"gcr.io/{self._get_project_or_default(self.project)}/"
                f"{image_name}"
            )
        if tag:
            image += f":{tag}"
        return image

    def create_resource(
        self,
        project: str | None = None,
        name: str | None = None,
        location: str | None = None,
        description: str | None = None,
        labels: dict[str, str] | None = None,
        format: Literal[
            "DOCKER", "MAVEN", "NPM", "PYTHON", "APT", "YUM"
        ] = "DOCKER",
        mode: Literal["STANDARD", "VIRTUAL", "REMOTE"] = "STANDARD",
    ) -> Response[GoogleArtifactRegistryResource]:
        from google.api_core.exceptions import NotFound
        from google.cloud import artifactregistry_v1
        from google.cloud.artifactregistry_v1 import Repository
        from google.protobuf.field_mask_pb2 import FieldMask

        repo_id = name or self.name
        if not repo_id:
            raise BadRequestError("Repository name is required.")

        location_input = location or self.location
        if not location_input:
            raise BadRequestError("Location is required.")

        project = project or self._get_project_or_default(project=self.project)
        client = artifactregistry_v1.ArtifactRegistryClient(
            credentials=self._get_credentials()
        )

        parent = f"projects/{project}/locations/{location_input}"
        repo_name = f"{parent}/repositories/{repo_id}"

        # Map format/mode to enums
        fmt_map = {
            "DOCKER": Repository.Format.DOCKER,
            "MAVEN": Repository.Format.MAVEN,
            "NPM": Repository.Format.NPM,
            "PYTHON": Repository.Format.PYTHON,
            "APT": Repository.Format.APT,
            "YUM": Repository.Format.YUM,
        }
        mode_map = {
            "STANDARD": Repository.Mode.STANDARD_REPOSITORY,
            "VIRTUAL": Repository.Mode.VIRTUAL_REPOSITORY,
            "REMOTE": Repository.Mode.REMOTE_REPOSITORY,
        }

        try:
            existing = client.get_repository(name=repo_name)
            update = Repository(
                name=repo_name,
                description=(
                    description
                    if description is not None
                    else existing.description
                ),
                labels=labels if labels is not None else dict(existing.labels),
            )
            mask_paths: list[str] = []
            if description is not None:
                mask_paths.append("description")
            if labels is not None:
                mask_paths.append("labels")

            if mask_paths:
                result = client.update_repository(
                    repository=update,
                    update_mask=FieldMask(paths=mask_paths),
                )
            else:
                result = existing
        except NotFound:
            create = Repository(
                format=fmt_map.get(format, Repository.Format.DOCKER),
                mode=mode_map.get(mode, Repository.Mode.STANDARD_REPOSITORY),
                description=description or "",
                labels=labels or {},
            )
            op = client.create_repository(
                parent=parent,
                repository_id=repo_id,
                repository=create,
            )
            result = op.result()

        login_server = f"{location_input}-docker.pkg.dev/{project}/{repo_id}"
        resource = GoogleArtifactRegistryResource(
            id=result.name,
            name=repo_id,
            location=location_input,
            login_server=login_server,
        )
        return Response(result=resource)

    def get_resource(
        self,
        project: str | None = None,
        name: str | None = None,
        location: str | None = None,
    ) -> Response[GoogleArtifactRegistryResource]:
        from google.api_core.exceptions import NotFound
        from google.cloud import artifactregistry_v1

        repo_id = name or self.name
        if not repo_id:
            raise BadRequestError("Repository name is required.")

        location_input = location or self.location
        if not location_input:
            raise BadRequestError("Location is required.")

        project = project or self._get_project_or_default(project=self.project)
        client = artifactregistry_v1.ArtifactRegistryClient(
            credentials=self._get_credentials()
        )

        parent = f"projects/{project}/locations/{location_input}"
        repo_name = f"{parent}/repositories/{repo_id}"

        try:
            repo = client.get_repository(name=repo_name)
        except NotFound:
            raise NotFoundError("Artifact Registry repository not found.")

        login_server = f"{location_input}-docker.pkg.dev/{project}/{repo_id}"
        result = GoogleArtifactRegistryResource(
            id=repo.name,
            name=repo_id,
            location=location_input,
            login_server=login_server,
        )
        return Response(result=result)

    def delete_resource(
        self,
        project: str | None = None,
        name: str | None = None,
        location: str | None = None,
    ) -> Response[None]:
        from google.api_core.exceptions import NotFound
        from google.cloud import artifactregistry_v1

        repo_id = name or self.name
        if not repo_id:
            raise BadRequestError("Repository name is required.")

        location_input = location or self.location
        if not location_input:
            raise BadRequestError("Location is required.")

        project = project or self._get_project_or_default(project=self.project)
        client = artifactregistry_v1.ArtifactRegistryClient(
            credentials=self._get_credentials()
        )

        parent = f"projects/{project}/locations/{location_input}"
        repo_name = f"{parent}/repositories/{repo_id}"

        try:
            client.get_repository(name=repo_name)
        except NotFound:
            raise NotFoundError("Artifact Registry repository not found.")

        op = client.delete_repository(name=repo_name)
        op.result()

        return Response(result=None)

    def push(self, image_name: str) -> Response[ContainerRegistryItem]:
        cmd = [
            "gcloud",
            "auth",
            "configure-docker",
        ]
        if self.name:
            cmd.append(self._artifact_registry_path())
        subprocess.run(
            cmd,
            shell=self._shell,
            check=True,
        )
        full_path = self._full_image_path(image_name)
        subprocess.run(
            ["docker", "tag", image_name, full_path],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "push", full_path], shell=self._shell, check=True
        )
        result = ContainerRegistryItem(
            image_name=image_name, image_uri=full_path
        )
        return Response(result=result)

    def pull(
        self, image_name: str, tag: str | None = None
    ) -> Response[ContainerRegistryItem]:
        full_path = self._full_image_path(image_name, tag)
        subprocess.run(
            [
                "gcloud",
                "auth",
                "configure-docker",
                self._artifact_registry_path(),
            ],
            shell=self._shell,
            check=True,
        )
        subprocess.run(
            ["docker", "pull", "--platform", self.platform, full_path],
            shell=self._shell,
            check=True,
        )
        result = ContainerRegistryItem(
            image_name=image_name, image_uri=full_path
        )
        return Response(result=result)

    def tag(
        self,
        image_name: str,
        tag: str,
        digest: str | None = None,
    ) -> Response[ContainerRegistryItemDigest]:
        self.__setup__()
        image_uri_base = self._full_image_path(image_name)
        if digest is None:
            digests = self.get_digests(image_name).result
            if not digests:
                raise BadRequestError(
                    f"No digests found for image '{image_name}'"
                )
            digest_entry = max(digests, key=lambda d: d.upload_time or 0)
            digest = digest_entry.digest

        subprocess.run(
            [
                "gcloud",
                "artifacts",
                "docker",
                "tags",
                "add",
                f"{image_uri_base}@{digest}",
                f"{image_uri_base}:{tag}",
            ],
            shell=self._shell,
            check=True,
        )
        result = ContainerRegistryItemDigest(
            image_uri=f"{image_uri_base}:{tag}",
            digest=digest,
            upload_time=None,
            image_size_bytes=None,
            tags=[tag],
        )
        return Response(result=result)

    def delete(
        self,
        image_name: str,
        digest: str | None = None,
        tag: str | None = None,
    ) -> Response[None]:
        full_path = self._full_image_path(image_name)

        if tag is not None and digest is None:
            subprocess.run(
                [
                    "gcloud",
                    "artifacts",
                    "docker",
                    "tags",
                    "delete",
                    f"{full_path}:{tag}",
                    "--quiet",
                ],
                shell=self._shell,
                check=True,
            )
            return Response(result=None)

        if digest is not None:
            target = f"{full_path}@{digest}"
        else:
            target = full_path

        subprocess.run(
            [
                "gcloud",
                "artifacts",
                "docker",
                "images",
                "delete",
                target,
                "--delete-tags",
                "--quiet",
            ],
            shell=self._shell,
            check=True,
        )
        return Response(result=None)

    def list_images(self) -> Response[list[ContainerRegistryItem]]:
        self.__setup__()
        parent = self._repository_path()
        results = self._client.list_docker_images(parent=parent)

        seen = set()
        items = []

        for image in results:
            parts = image.uri.split("/")
            image_path = parts[-1]
            base_name, digest = image_path.split("@")

            if base_name not in seen:
                seen.add(base_name)
                items.append(
                    ContainerRegistryItem(
                        image_name=base_name,
                        image_uri=image.uri.split("@")[0],
                    )
                )

        result = items
        return Response(result=result)

    def get_digests(
        self, image_name: str
    ) -> Response[list[ContainerRegistryItemDigest]]:
        self.__setup__()
        parent = self._repository_path()
        results = self._client.list_docker_images(parent=parent)
        digests = []

        for image in results:
            parts = image.name.split("/")
            image_path = parts[-1]
            base_name, digest = image_path.split("@")
            if base_name == image_name:
                digests.append(
                    ContainerRegistryItemDigest(
                        image_uri=image.uri,
                        digest=digest,
                        upload_time=image.upload_time.timestamp(),
                        image_size_bytes=image.image_size_bytes,
                        tags=list(image.tags) if image.tags else None,
                    )
                )

        result = digests
        return Response(result=result)

    async def alist_images(self) -> Response[list[ContainerRegistryItem]]:
        await self.__asetup__()
        parent = self._repository_path()
        results = await self._aclient.list_docker_images(parent=parent)

        seen = set()
        items = []

        async for image in results:
            parts = image.uri.split("/")
            image_path = parts[-1]
            base_name, digest = image_path.split("@")

            if base_name not in seen:
                seen.add(base_name)
                items.append(
                    ContainerRegistryItem(
                        image_name=base_name,
                        image_uri="/".join(parts[:-1]),
                    )
                )

        result = items
        return Response(result=result)

    async def aget_digests(
        self, image_name: str
    ) -> Response[list[ContainerRegistryItemDigest]]:
        await self.__asetup__()
        parent = self._repository_path()
        results = await self._aclient.list_docker_images(parent=parent)
        digests = []

        async for image in results:
            parts = image.name.split("/")
            image_path = parts[-1]
            base_name, digest = image_path.split("@")
            if base_name == image_name:
                digests.append(
                    ContainerRegistryItemDigest(
                        image_uri=image.uri,
                        digest=digest,
                        upload_time=image.upload_time.timestamp(),
                        image_size_bytes=image.image_size_bytes,
                        tags=list(image.tags) if image.tags else None,
                    )
                )

        result = digests
        return Response(result=result)

    def close(self) -> Response[None]:
        self._init = False
        return Response(result=None)

    async def aclose(self) -> Response[None]:
        self._ainit = False
        return Response(result=None)



from typing import Any

from common.secrets import get_secrets
from x8.storage.document_store import DocumentStore

secrets = get_secrets()


class DocumentStoreProvider:
    AMAZON_DYNAMODB = "amazon_dynamodb"
    AZURE_COSMOS_DB = "azure_cosmos_db"
    GOOGLE_FIRESTORE = "google_firestore"
    MONGODB = "mongodb"
    POSTGRESQL = "postgresql"
    REDIS = "redis"
    SQLITE = "sqlite"
    MEMORY = "memory"


provider_parameters: dict[str, dict[str, Any]] = {
    DocumentStoreProvider.AMAZON_DYNAMODB: {
        "region_name": secrets["storage-document-amazon-dynamodb-region-name"],
        "aws_access_key_id": secrets[
            "storage-document-amazon-dynamodb-aws-access-key-id"
        ],
        "aws_secret_access_key": secrets[
            "storage-document-amazon-dynamodb-aws-secret-access-key"
        ],
    },
    DocumentStoreProvider.AZURE_COSMOS_DB: {
        "endpoint": secrets["storage-document-azure-cosmos-db-endpoint"],
        "access_key": secrets["storage-document-azure-cosmos-db-access-key"],
        "database": "test",
        "suppress_fields": ["_rid", "_ts", "_self", "_attachments"],
    },
    DocumentStoreProvider.GOOGLE_FIRESTORE: {
        "project": secrets["storage-document-google-firestore-project"],
        "database": "test",
    },
    DocumentStoreProvider.MONGODB: {
        "uri": secrets["storage-document-mongodb-uri"],
        "database": "test",
        "suppress_fields": ["_id"],
    },
    DocumentStoreProvider.POSTGRESQL: {
        "connection_string": secrets[
            "storage-document-postgresql-connection-string"
        ],
        "id_column": "id",
        "value_column": "value",
    },
    DocumentStoreProvider.REDIS: {
        "host": secrets["storage-document-redis-host"],
        "port": secrets["storage-document-redis-port"],
        "password": secrets["storage-document-redis-password"],
        "field_types": {
            "bool": "boolean",
            "obj": "object",
            "obj.nint": "number",
        },
    },
    DocumentStoreProvider.SQLITE: {},
    DocumentStoreProvider.MEMORY: {},
}


def get_component(provider_type: str, collection: str = "test"):
    component = DocumentStore(
        collection=collection,
        __provider__=dict(
            type=provider_type,
            parameters=provider_parameters[provider_type],
        ),
    )
    return component



from __future__ import annotations

import base64
import json
from typing import Any

import boto3
from botocore.exceptions import ClientError
from x8.core import Context, DataModel, NCall, Operation, Provider, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError
from x8.messaging._common import (
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingOperation,
    MessagingOperationParser,
    QueueConfig,
    QueueInfo,
)

from .._feature import QueueFeature


class AmazonSQS(Provider):
    queue: str | None
    region: str | None
    profile_name: str | None
    aws_access_key_id: str | None
    aws_secret_access_key: str | None
    aws_session_token: str | None
    nparams: dict[str, Any]

    _sqs_client: Any
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _queue_url_cache: dict[str, str]

    def __init__(
        self,
        queue: str | None = None,
        region: str | None = None,
        profile_name: str | None = None,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        aws_session_token: str | None = None,
        nparams: dict[str, Any] = dict(),
        **kwargs: Any,
    ):
        """Initialize.

        Args:
            queue:
                SQS queue name.
            region_name:
                AWS region name.
            profile_name:
                AWS profile name.
            aws_access_key_id:
                AWS access key ID.
            aws_secret_access_key:
                AWS secret access key.
            aws_session_token:
                AWS session token.
            nparams:
                Native parameters to SQS client.
        """
        self.queue = queue
        self.region = region
        self.profile_name = profile_name
        self.access_key_id = aws_access_key_id
        self.secret_access_key = aws_secret_access_key
        self.session_token = aws_session_token
        self.nparams = nparams

        self._sqs_client = None
        self._op_converter = OperationConverter()
        self._result_converter = ResultConverter()
        self._queue_url_cache = dict()
        super().__init__(**kwargs)

    def __supports__(self, feature: str) -> bool:
        return feature not in [QueueFeature.BUILTIN_DLQ]

    def __setup__(self, context: Context | None = None) -> None:
        if self._sqs_client is not None:
            return

        session_kwargs = {}
        if self.profile_name:
            session_kwargs["profile_name"] = self.profile_name

        client_kwargs = {"region_name": self.region, **self.nparams}

        if self.access_key_id and self.secret_access_key:
            client_kwargs["aws_access_key_id"] = self.access_key_id
            client_kwargs["aws_secret_access_key"] = self.secret_access_key
            if self.session_token:
                client_kwargs["aws_session_token"] = self.session_token

        session = boto3.Session(**session_kwargs)
        self._sqs_client = session.client("sqs", **client_kwargs)

    def _get_queue_name(
        self, op_parser: MessagingOperationParser
    ) -> str | None:
        queue_name = op_parser.get_queue()
        queue_name = queue_name or self.queue or self.__component__.queue
        return queue_name

    def _get_queue_url(self, op_parser: MessagingOperationParser) -> str:
        queue_name = self._get_queue_name(op_parser)
        if queue_name is None:
            raise BadRequestError("Queue name is required")

        if queue_name in self._queue_url_cache:
            return self._queue_url_cache[queue_name]

        try:
            response = self._sqs_client.get_queue_url(QueueName=queue_name)
            self._queue_url_cache[queue_name] = response["QueueUrl"]
            return response["QueueUrl"]
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                raise BadRequestError(f"Queue {queue_name} does not exist")
            raise

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        ncall = self._get_ncall(
            op_parser,
            ClientHelper(
                self._sqs_client,
                self._op_converter,
            ),
        )
        if ncall is None:
            return super().__run__(
                operation=operation,
                context=context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(nresult, op_parser)
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        client_helper: ClientHelper,
    ) -> NCall | None:
        call = None
        nargs = op_parser.get_nargs()

        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            args: dict = {
                "queue_name": self._get_queue_name(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "config": op_parser.get_queue_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.create_queue,
                args,
                None,
            )
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "config": op_parser.get_queue_config(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.update_queue,
                args,
                None,
            )
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "where_exists": op_parser.get_where_exists(),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.drop_queue,
                args,
                None,
            )
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            args = {}
            call = NCall(
                self._sqs_client.list_queues,
                args,
                nargs,
            )
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.has_queue,
                args,
                None,
            )
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            args = {
                "queue_name": self._get_queue_name(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.get_queue,
                args,
                None,
            )
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = self._op_converter.convert_put(
                self._get_queue_url(op_parser),
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            call = NCall(
                self._sqs_client.send_message,
                args,
                nargs,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = {
                "batch": op_parser.get_batch(),
                "queue_url": self._get_queue_url(op_parser),
            }
            call = NCall(
                client_helper.batch,
                args,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            args = self._op_converter.convert_pull(
                self._get_queue_url(op_parser),
                op_parser.get_pull_config(),
            )
            call = NCall(
                self._sqs_client.receive_message,
                args,
                nargs,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = {
                "QueueUrl": self._get_queue_url(op_parser),
                "ReceiptHandle": op_parser.get_key().nref,
            }
            call = NCall(
                self._sqs_client.delete_message,
                args,
                nargs,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = {
                "QueueUrl": self._get_queue_url(op_parser),
                "ReceiptHandle": op_parser.get_key().nref,
                "VisibilityTimeout": 0,
            }
            call = NCall(
                self._sqs_client.change_message_visibility,
                args,
                nargs,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = {
                "QueueUrl": self._get_queue_url(op_parser),
                "ReceiptHandle": op_parser.get_key().nref,
                "VisibilityTimeout": op_parser.get_timeout(),
            }
            call = NCall(
                self._sqs_client.change_message_visibility,
                args,
                nargs,
            )
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            args = {
                "queue_url": self._get_queue_url(op_parser),
                "nargs": nargs,
            }
            call = NCall(
                client_helper.purge,
                args,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {}
            call = NCall(client_helper.close, args)

        return call

    def _convert_nresult(
        self,
        nresult: Any,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result_converter = self._result_converter
        result: Any = None

        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            result = nresult
            queue_name = self._get_queue_name(op_parser)
            if queue_name:
                self._queue_url_cache.pop(queue_name, None)
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            result = None
            queue_name = self._get_queue_name(op_parser)
            if queue_name:
                self._queue_url_cache.pop(queue_name, None)
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            result = result_converter.convert_list_queues(nresult)
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            result = nresult
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            result = nresult
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = result_converter.convert_pull(nresult)
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            result = MessageItem(key=op_parser.get_key())
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            result = None
        return result


class ClientHelper:
    sqs_client: Any
    op_converter: OperationConverter

    def __init__(
        self,
        sqs_client: Any,
        op_converter: OperationConverter,
    ):
        self.sqs_client = sqs_client
        self.op_converter = op_converter

    def create_queue(
        self,
        queue_name: str,
        where_exists: bool | None,
        config: QueueConfig | None,
        nargs: Any,
    ) -> None:
        try:
            NCall(
                self.sqs_client.get_queue_url,
                {"QueueName": queue_name},
            ).invoke()
            if where_exists is False:
                raise ConflictError(f"Queue {queue_name} already exists")
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                attributes = self.op_converter.convert_queue_attributes(
                    config=config,
                )
                NCall(
                    self.sqs_client.create_queue,
                    {
                        "QueueName": queue_name,
                        "Attributes": attributes,
                    },
                    nargs,
                ).invoke()
            else:
                raise

    def update_queue(
        self,
        queue_name: str,
        config: QueueConfig | None,
        nargs: Any,
    ) -> None:
        try:
            queue_url = NCall(
                self.sqs_client.get_queue_url,
                {"QueueName": queue_name},
            ).invoke()["QueueUrl"]
            attributes = self.op_converter.convert_queue_attributes(
                config=config
            )
            NCall(
                self.sqs_client.set_queue_attributes,
                {
                    "QueueUrl": queue_url,
                    "Attributes": attributes,
                },
                nargs,
            ).invoke()
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                raise NotFoundError(f"Queue {queue_name} does not exist")
            else:
                raise

    def drop_queue(
        self, queue_name: str, where_exists: bool | None, nargs: Any
    ) -> None:
        try:
            queue_url = NCall(
                self.sqs_client.get_queue_url,
                {"QueueName": queue_name},
            ).invoke()["QueueUrl"]
            NCall(
                self.sqs_client.delete_queue,
                {"QueueUrl": queue_url},
                nargs,
            ).invoke()
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                if where_exists is True:
                    raise NotFoundError(f"Queue {queue_name} does not exist")
            else:
                raise

    def has_queue(
        self,
        queue_name: str,
        nargs: Any,
    ) -> bool:
        try:
            NCall(
                self.sqs_client.get_queue_url,
                {"QueueName": queue_name},
                nargs,
            ).invoke()
            return True
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                return False
            raise

    def get_queue(
        self,
        queue_name: str,
        nargs: Any,
    ) -> QueueInfo:
        try:
            res = NCall(
                self.sqs_client.get_queue_url,
                {"QueueName": queue_name},
                nargs,
            ).invoke()
            queue_url = res["QueueUrl"]
            res = NCall(
                self.sqs_client.get_queue_attributes,
                {
                    "QueueUrl": queue_url,
                    "AttributeNames": [
                        "All",
                    ],
                },
                nargs,
            ).invoke()
            attributes = res["Attributes"]
            fifo = attributes.get("FifoQueue", "false").lower() == "true"
            queue_arn = attributes.get("QueueArn")
            max_receive_count = None
            dlq_nref = None
            if "RedrivePolicy" in attributes:
                redrive_policy = json.loads(attributes["RedrivePolicy"])
                dlq_nref = redrive_policy.get("deadLetterTargetArn")
                if "maxReceiveCount" in redrive_policy:
                    max_receive_count = int(
                        redrive_policy.get("maxReceiveCount", 0)
                    )
            return QueueInfo(
                name=queue_name,
                active_message_count=int(
                    attributes.get("ApproximateNumberOfMessages", 0)
                ),
                inflight_message_count=int(
                    attributes.get("ApproximateNumberOfMessagesNotVisible", 0)
                ),
                scheduled_message_count=int(
                    attributes.get("ApproximateNumberOfMessagesDelayed", 0)
                ),
                config=QueueConfig(
                    visibility_timeout=int(
                        attributes.get("VisibilityTimeout", 0)
                    ),
                    ttl=int(attributes.get("MessageRetentionPeriod", 0)),
                    max_delivery_count=max_receive_count,
                    dlq_nref=dlq_nref,
                    fifo=fifo,
                ),
                nref=queue_arn,
            )
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.NonExistentQueue"
            ):
                raise NotFoundError(f"Queue {queue_name} does not exist")
            raise

    def batch(
        self,
        batch: MessageBatch,
        queue_url: str,
    ) -> list:
        entries = []
        for i, operation in enumerate(batch.operations):
            op_parser = MessagingOperationParser(operation)
            args = self.op_converter.convert_put(
                queue_url,
                op_parser.get_value(),
                op_parser.get_metadata(),
                op_parser.get_properties(),
                op_parser.get_put_config(),
            )
            entry = {
                "Id": str(i),
                "MessageBody": args["MessageBody"],
            }

            if "MessageAttributes" in args:
                entry["MessageAttributes"] = args["MessageAttributes"]

            if "DelaySeconds" in args:
                entry["DelaySeconds"] = args["DelaySeconds"]

            if "MessageGroupId" in args:
                entry["MessageGroupId"] = args["MessageGroupId"]

            if "MessageDeduplicationId" in args:
                entry["MessageDeduplicationId"] = args[
                    "MessageDeduplicationId"
                ]

            entries.append(entry)

        response = NCall(
            self.sqs_client.send_message_batch,
            {"QueueUrl": queue_url, "Entries": entries},
        ).invoke()

        message_ids = []
        for msg in response.get("Successful", []):
            message_ids.append(msg.get("MessageId"))

        if "Failed" in response and response["Failed"]:
            for failed in response["Failed"]:
                raise BadRequestError(
                    (
                        f"Failed to send message {failed['Id']}: "
                        f"{failed['Message']}"
                    )
                )
        return message_ids

    def purge(
        self,
        queue_url: str,
        nargs: Any,
    ) -> None:
        try:
            NCall(
                self.sqs_client.purge_queue,
                {"QueueUrl": queue_url},
                nargs,
            ).invoke()
        except ClientError as e:
            if (
                e.response["Error"]["Code"]
                == "AWS.SimpleQueueService.PurgeQueueInProgress"
            ):
                pass

    def close(self) -> Any:
        return None


class OperationConverter:
    def convert_queue_attributes(
        self,
        config: QueueConfig | None,
    ) -> dict:
        attributes = {}
        if config:
            if config.visibility_timeout:
                attributes["VisibilityTimeout"] = str(
                    int(config.visibility_timeout)
                )
            if config.ttl:
                attributes["MessageRetentionPeriod"] = str(int(config.ttl))
            if config.fifo:
                attributes["FifoQueue"] = "true"
            if config.dlq_nref and config.max_delivery_count:
                redrive_policy = {
                    "deadLetterTargetArn": config.dlq_nref,
                    "maxReceiveCount": str(config.max_delivery_count),
                }
                attributes["RedrivePolicy"] = json.dumps(redrive_policy)
            if config.nconfig:
                attributes.update(config.nconfig)
        return attributes

    def convert_put(
        self,
        queue_url: str,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
        config: MessagePutConfig | None,
    ) -> dict:
        message_body: str = ""
        message_attributes: dict = {}
        if isinstance(value, str):
            message_body = value
            content_type = "text/plain"
        elif isinstance(value, bytes):
            message_body = base64.b64encode(value).decode("utf-8")
            content_type = "application/octet-stream"
        elif isinstance(value, dict):
            message_body = json.dumps(value)
            content_type = "application/json"
        elif isinstance(value, DataModel):
            message_body = value.to_json()
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")

        message_attributes["content_type"] = {
            "DataType": "String",
            "StringValue": content_type,
        }

        if metadata:
            for key, val in metadata.items():
                if isinstance(val, str):
                    message_attributes[key] = {
                        "DataType": "String",
                        "StringValue": val,
                    }
                elif isinstance(val, (int, float)):
                    message_attributes[key] = {
                        "DataType": "Number",
                        "StringValue": str(val),
                    }
                elif isinstance(val, bytes):
                    message_attributes[key] = {
                        "DataType": "Binary",
                        "BinaryValue": val,
                    }
                else:
                    message_attributes[key] = {
                        "DataType": "String",
                        "StringValue": str(val),
                    }

        message_deduplication_id = None
        message_group_id = None
        if properties:
            if properties.message_id:
                message_attributes["message_id"] = {
                    "DataType": "String",
                    "StringValue": properties.message_id,
                }
                message_deduplication_id = properties.message_id

            if properties.group_id:
                message_attributes["group_id"] = {
                    "DataType": "String",
                    "StringValue": properties.group_id,
                }
                message_group_id = properties.group_id

        args: dict = {
            "QueueUrl": queue_url,
            "MessageBody": message_body,
        }

        if message_attributes:
            args["MessageAttributes"] = message_attributes

        if message_group_id and queue_url.endswith(".fifo"):
            args["MessageGroupId"] = message_group_id

        if message_deduplication_id and queue_url.endswith(".fifo"):
            args["MessageDeduplicationId"] = message_deduplication_id

        if config and config.delay is not None:
            args["DelaySeconds"] = min(900, int(config.delay))
        return args

    def convert_pull(
        self, queue_url: str, config: MessagePullConfig | None
    ) -> dict:
        args: dict = {
            "QueueUrl": queue_url,
            "MessageAttributeNames": ["All"],
            "AttributeNames": ["All"],
        }
        if config:
            if config.max_wait_time is not None:
                args["WaitTimeSeconds"] = int(config.max_wait_time)

            if config.max_count is not None:
                args["MaxNumberOfMessages"] = min(10, config.max_count)

            if config.visibility_timeout is not None:
                args["VisibilityTimeout"] = int(config.visibility_timeout)

        return args


class ResultConverter:
    def _convert_key(self, message: dict) -> MessageKey:
        return MessageKey(nref=message.get("ReceiptHandle"))

    def _convert_value(self, message: dict) -> MessageValueType:
        message_attributes = message.get("MessageAttributes", {})
        content_type = message_attributes.get(
            "content_type", {"StringValue": "text/plain"}
        ).get("StringValue", "text/plain")

        body = message.get("Body", "")

        if content_type == "application/octet-stream":
            return base64.b64decode(body)
        elif content_type == "application/json":
            try:
                return json.loads(body)
            except json.JSONDecodeError:
                return body

        return body

    def _convert_metadata(self, message: dict) -> dict | None:
        message_attributes = message.get("MessageAttributes", {})
        if not message_attributes:
            return None

        metadata = {}
        for key, attr in message_attributes.items():
            if key in ["content_type", "message_id", "group_id"]:
                continue

            if attr.get("DataType") == "String":
                metadata[key] = attr.get("StringValue")
            elif attr.get("DataType") == "Number":
                try:
                    metadata[key] = float(attr.get("StringValue"))
                    if metadata[key].is_integer():
                        metadata[key] = int(metadata[key])
                except (ValueError, TypeError):
                    metadata[key] = attr.get("StringValue")
            elif attr.get("DataType") == "Binary":
                metadata[key] = attr.get("BinaryValue")
            else:
                metadata[key] = attr.get("StringValue")

        return metadata if metadata else None

    def _convert_properties(self, message: dict) -> MessageProperties:
        message_attributes = message.get("MessageAttributes", {})
        attributes = message.get("Attributes", {})

        message_id = None
        if "message_id" in message_attributes:
            message_id = message_attributes["message_id"].get("StringValue")

        group_id = None
        if "group_id" in message_attributes:
            group_id = message_attributes["group_id"].get("StringValue")

        content_type = "text/plain"
        if "content_type" in message_attributes:
            content_type = message_attributes["content_type"].get(
                "StringValue", "text/plain"
            )

        enqueued_time = None
        if "SentTimestamp" in attributes:
            try:
                enqueued_time = float(attributes["SentTimestamp"]) / 1000.0
            except (ValueError, TypeError):
                pass

        # Get other properties
        return MessageProperties(
            message_id=message_id or message.get("MessageId"),
            content_type=content_type,
            group_id=group_id,
            enqueued_time=enqueued_time,
            delivery_count=(
                int(attributes.get("ApproximateReceiveCount", 1))
                if attributes
                else 1
            ),
        )

    def _convert_message(
        self,
        message: dict,
    ) -> MessageItem:
        return MessageItem(
            key=self._convert_key(message=message),
            value=self._convert_value(message=message),
            metadata=self._convert_metadata(message=message),
            properties=self._convert_properties(message=message),
        )

    def convert_list_queues(self, nresult: dict) -> list[str]:
        queues = []
        for queue_url in nresult.get("QueueUrls", []):
            queue_name = queue_url.split("/")[-1]
            queues.append(queue_name)
        return queues

    def convert_pull(self, nresult: dict) -> list[MessageItem]:
        if not nresult or "Messages" not in nresult:
            return []

        result = []
        for message in nresult["Messages"]:
            result.append(self._convert_message(message))
        return result



from common.sync_and_async_client import SyncAndAsyncClient

from ._providers import get_component


class QueueSyncAndAsyncClient(SyncAndAsyncClient):
    def __init__(self, provider_type: str, async_call: bool):
        self.client = get_component(provider_type)
        self.async_call = async_call
        self.provider_type = provider_type

    async def create_queue(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def update_queue(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def drop_queue(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def list_queues(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def has_queue(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def get_queue(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def put(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def batch(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def pull(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def ack(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def nack(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def extend(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def purge(self, **kwargs):
        return await self._execute_method(**kwargs)

    async def close(self, **kwargs):
        return await self._execute_method(**kwargs)



from x8.core import OperationParser

from ._models import (
    MessageBatch,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    QueueConfig,
    SubscriptionConfig,
    TopicConfig,
)
from ._operation import MessagingOperation


class MessagingOperationParser(OperationParser):
    def get_value(self) -> MessageValueType:
        return self.get_arg("value")

    def get_metadata(self) -> dict | None:
        return self.get_arg("metadata")

    def get_properties(self) -> MessageProperties | None:
        properties = self.get_arg("properties")
        if isinstance(properties, dict):
            return MessageProperties.from_dict(properties)
        return properties

    def get_put_config(self) -> MessagePutConfig | None:
        config = self.get_arg("config")
        if isinstance(config, dict):
            return MessagePutConfig.from_dict(config)
        return config

    def get_pull_config(self) -> MessagePullConfig | None:
        config = self.get_arg("config")
        if isinstance(config, dict):
            return MessagePullConfig.from_dict(config)
        return config

    def get_batch(self) -> MessageBatch:
        batch = self.get_arg("batch")
        if isinstance(batch, dict):
            return MessageBatch.from_dict(batch)
        return batch

    def get_key(self) -> MessageKey:
        key = self.get_arg("key")
        if isinstance(key, dict):
            return MessageKey.from_dict(key)
        return key

    def get_timeout(self) -> int | None:
        return self.get_arg("timeout")

    def get_queue(self) -> str | None:
        return self.get_arg("queue")

    def get_topic(self) -> str | None:
        return self.get_arg("topic")

    def get_subscription(self) -> str | None:
        return self.get_arg("subscription")

    def get_queue_config(self) -> QueueConfig | None:
        config = self.get_arg("config")
        if isinstance(config, dict):
            return QueueConfig.from_dict(config)
        return config

    def get_topic_config(self) -> TopicConfig | None:
        config = self.get_arg("config")
        if isinstance(config, dict):
            return TopicConfig.from_dict(config)
        return config

    def get_subscription_config(self) -> SubscriptionConfig | None:
        config = self.get_arg("config")
        if isinstance(config, dict):
            return SubscriptionConfig.from_dict(config)
        return config

    def is_resource_op(self) -> bool:
        return self.get_op_name() in [
            MessagingOperation.CREATE_QUEUE,
            MessagingOperation.UPDATE_QUEUE,
            MessagingOperation.DROP_QUEUE,
            MessagingOperation.LIST_QUEUES,
            MessagingOperation.HAS_QUEUE,
            MessagingOperation.GET_QUEUE,
            MessagingOperation.CREATE_TOPIC,
            MessagingOperation.UPDATE_TOPIC,
            MessagingOperation.DROP_TOPIC,
            MessagingOperation.LIST_TOPICS,
            MessagingOperation.HAS_TOPIC,
            MessagingOperation.GET_TOPIC,
            MessagingOperation.CREATE_SUBSCRIPTION,
            MessagingOperation.UPDATE_SUBSCRIPTION,
            MessagingOperation.DROP_SUBSCRIPTION,
            MessagingOperation.LIST_SUBSCRIPTIONS,
            MessagingOperation.HAS_SUBSCRIPTION,
            MessagingOperation.GET_SUBSCRIPTION,
            MessagingOperation.CLOSE,
        ]

    def is_sender_op(self):
        return self.get_op_name() in [
            MessagingOperation.PUT,
            MessagingOperation.BATCH,
        ]

    def is_receiver_op(self):
        return self.get_op_name() in [
            MessagingOperation.PULL,
            MessagingOperation.ACK,
            MessagingOperation.NACK,
            MessagingOperation.EXTEND,
            MessagingOperation.PURGE,
        ]



from botocore.exceptions import ClientError


def ensure_launch_template(
    ec2_client,
    launch_template_name: str,
    ami_id: str,
    instance_type: str,
    security_group_id: str,
    user_data: str,
    instance_profile_arn: str | None = None,
    keypair_name: str | None = None,
    subnet_id: str | None = None,
) -> tuple[str, str]:
    try:
        lt = ec2_client.describe_launch_templates(
            LaunchTemplateNames=[launch_template_name]
        )["LaunchTemplates"][0]
        return lt["LaunchTemplateId"], "$Latest"
    except ClientError as e:
        if (
            e.response["Error"]["Code"]
            != "InvalidLaunchTemplateName.NotFoundException"
        ):
            raise

    # Create fresh
    launch_template_data = {
        "ImageId": ami_id,
        "InstanceType": instance_type,
        "SecurityGroupIds": [security_group_id],
        "UserData": user_data,
    }

    if instance_profile_arn:
        launch_template_data["IamInstanceProfile"] = {
            "Arn": instance_profile_arn  # type: ignore
        }

    if keypair_name:
        launch_template_data["KeyName"] = keypair_name

    if subnet_id:
        launch_template_data["SubnetId"] = subnet_id

    resp = ec2_client.create_launch_template(
        LaunchTemplateName=launch_template_name,
        LaunchTemplateData=launch_template_data,
    )
    lt = resp["LaunchTemplate"]
    print(f"Created launch template: {launch_template_name}")
    return lt["LaunchTemplateId"], "$Latest"


def delete_launch_template(
    ec2_client,
    launch_template_name: str,
) -> None:
    try:
        ec2_client.delete_launch_template(
            LaunchTemplateName=launch_template_name
        )
    except ClientError as e:
        code = e.response.get("Error", {}).get("Code", "")
        if code in (
            "InvalidLaunchTemplateId.NotFound",
            "InvalidLaunchTemplateName.NotFoundException",
        ):
            return

        else:
            raise


def auto_detect_network_config(
    ec2_client,
    vpc_id: str | None = None,
    subnet_ids: list[str] = [],
) -> tuple[str | None, list[str] | None]:
    """Auto-detect VPC and subnet configuration."""
    vpc_id = vpc_id
    subnet_ids = subnet_ids
    try:
        if not vpc_id:
            # Get default VPC
            vpcs = ec2_client.describe_vpcs(
                Filters=[{"Name": "is-default", "Values": ["true"]}]
            )
            if vpcs["Vpcs"]:
                vpc_id = vpcs["Vpcs"][0]["VpcId"]

        if not subnet_ids and vpc_id:
            # Get subnets in the detected VPC
            subnets = ec2_client.describe_subnets(
                Filters=[{"Name": "vpc-id", "Values": [vpc_id]}]
            )
            subnet_ids = [subnet["SubnetId"] for subnet in subnets["Subnets"]]

    except Exception as e:
        print(f"Warning: Could not auto-detect network config: {e}")

    return vpc_id, subnet_ids



objects = [
    {
        "id": "test1.txt",
        "value": b"Hello World One",
    },
    {
        "id": "test2.txt",
        "value": b"Hello World Two",
        "properties": {
            "content_type": "text/plain",
            "content_disposition": "inline",
            "content_encoding": "none",
            "content_language": "en-US",
            "cache_control": "no-cache",
        },
    },
    {
        "id": "test3.txt",
        "value": b"Hello World Three",
        "metadata": {
            "str": "value3",
            "int": "30",
            "bool": "false",
            "arr": "tag1, tag2",
        },
    },
    {
        "id": "/c/d/test4.txt",
        "value": b"Hello World Four",
        "properties": {
            "content_type": "text/plain",
            "content_disposition": "inline",
            "content_encoding": "none",
            "content_language": "en-US",
            "cache_control": "no-cache",
        },
        "metadata": {
            "str": "value4",
            "int": "40",
            "bool": "true",
            "arr": "tag1",
        },
    },
    {
        "id": "test5.jpg",
        "file": "test-image.jpg",
        "properties": {
            "content_type": "image/jpeg",
            "storage_class": "hot",
        },
        "metadata": {
            "str": "value5",
            "int": "50",
            "bool": "true",
            "arr": "tag2",
        },
    },
    {
        "id": "test6.json",
        "file": "test-json.json",
        "properties": {
            "content_type": "application/json",
            "storage_class": "cool",
        },
        "metadata": {
            "str": "value6",
            "int": "60",
            "bool": "false",
            "arr": "tag1",
        },
    },
    {
        "id": "test7.txt",
        "stream": "test-text.txt",
        "properties": {
            "content_type": "text/plain",
        },
        "metadata": {
            "str": "value7",
            "int": "70",
            "bool": "true",
            "arr": "tag2",
        },
    },
    {
        "id": "test8.mp4",
        "stream": "test-video.mp4",
        "properties": {
            "content_type": "video/mp4",
        },
        "metadata": {
            "str": "value8",
            "int": "80",
            "bool": "false",
            "arr": "tag1, tag2",
        },
    },
]

query_objects = [
    {
        "id": "test00.txt",
        "value": b"value0",
    },
    {
        "id": "test01.txt",
        "value": b"value1",
    },
    {
        "id": "data/test02.txt",
        "value": b"value2",
    },
    {
        "id": "data/test03.txt",
        "value": b"value3",
    },
    {
        "id": "data/ab/test04.txt",
        "value": b"value4",
    },
    {
        "id": "data/ab/test05.txt",
        "value": b"value5",
    },
    {
        "id": "data/cd/test06.txt",
        "value": b"value6",
    },
    {
        "id": "data/xy/test07.txt",
        "value": b"value7",
    },
    {
        "id": "data/xy/test08.txt",
        "value": b"value8",
    },
    {
        "id": "abc/test09.txt",
        "value": b"value9",
    },
    {
        "id": "abc/test10.txt",
        "value": b"value10",
    },
    {
        "id": "tzyx/test13.txt",
        "value": b"value11",
    },
    {
        "id": "tzyx/test14.txt",
        "value": b"value12",
    },
    {
        "id": "aaa.txt",
        "value": b"value13",
    },
]



from __future__ import annotations

import json
import sqlite3
import time
import uuid
from typing import Any

from x8.core import Context, DataModel, NCall, Operation, Provider, Response
from x8.core.exceptions import BadRequestError, ConflictError, NotFoundError
from x8.core.time import Time

from ._models import (
    DEFAULT_SUBSCRIPTION_NAME,
    MessageBatch,
    MessageItem,
    MessageKey,
    MessageProperties,
    MessagePullConfig,
    MessagePutConfig,
    MessageValueType,
    MessagingMode,
    QueueConfig,
    QueueInfo,
    SubscriptionConfig,
    SubscriptionInfo,
    TopicConfig,
    TopicInfo,
)
from ._operation import MessagingOperation
from ._operation_parser import MessagingOperationParser

DEFAULT_TTL = 7 * 24 * 60 * 60
DEFAULT_VISIBILITY_TIMEOUT = 30


class SQLiteBase(Provider):
    mode: MessagingMode

    queue: str | None
    topic: str | None
    subscription: str | None

    database: str
    message_table: str
    metadata_table: str
    poll_interval: float
    nparams: dict[str, Any]

    _client: Any
    _client_helper: Any
    _op_converter: OperationConverter
    _result_converter: ResultConverter
    _topic_config_cache: dict[str, TopicConfig] = {}
    _subscription_config_cache: dict[
        str, dict[str, SubscriptionConfig | None]
    ] = {}

    def __init__(
        self,
        mode: MessagingMode = MessagingMode.QUEUE,
        database: str = ":memory:",
        queue: str | None = None,
        topic: str | None = None,
        subscription: str | None = None,
        message_table: str = "message",
        metadata_table: str = "metadata",
        poll_interval: float = 0.5,
        nparams: dict[str, Any] = dict(),
        **kwargs,
    ):
        """Initialize.

        Args:
            mode:
                Messaging mode.
            database:
                SQLite database. Defaults to ":memory:".
            queue:
                Queue name.
            topic:
                Topic name.
            subscription:
                Subscription name.
            message_table:
                SQLite table name for messages.
            metadata_table:
                SQLite table name for metadata.
            poll_interval:
                Poll interval in seconds. Defaults to 0.5.
            nparams:
                Native parameters to SQLite client.
        """
        self.mode = mode
        self.database = database
        self.queue = queue
        self.topic = topic
        self.subscription = subscription
        self.message_table = message_table
        self.metadata_table = metadata_table
        self.poll_interval = poll_interval
        self.nparams = nparams

        self._client = None
        self._client_helper = None
        self._op_converter = OperationConverter(
            self.message_table,
            self.metadata_table,
        )
        self._result_converter = ResultConverter()
        self._topic_config_cache = dict()
        self._subscription_config_cache = dict()
        if self.mode == MessagingMode.QUEUE:
            self.topic = self.queue
            self.subscription = DEFAULT_SUBSCRIPTION_NAME
        super().__init__(**kwargs)

    def __setup__(
        self,
        context: Context | None = None,
    ) -> None:
        if self._client is not None:
            return

        self._client = sqlite3.connect(
            database=self.database,
            check_same_thread=False,
            **self.nparams,
        )
        self._client_helper = ClientHelper(
            self._client,
            self._op_converter,
            self._result_converter,
            self.mode,
            self.poll_interval,
            self._subscription_config_cache,
        )
        self._create_tables_if_needed()

    def _create_tables_if_needed(self, **kwargs: Any) -> None:
        cursor = self._client.cursor()
        rows = cursor.execute(
            """SELECT name FROM sqlite_master WHERE type='table'
                AND name=?""",
            (self.message_table,),
        ).fetchall()
        if len(rows) == 0:
            cursor.execute(
                f"""
                CREATE TABLE {self.message_table} (
                    id TEXT,
                    topic TEXT,
                    subscription TEXT,
                    value BLOB,
                    metadata TEXT,
                    message_id TEXT,
                    group_id TEXT,
                    content_type TEXT,
                    enqueued_time REAL,
                    delivery_count INTEGER,
                    lock_until_time REAL,
                    lock_token TEXT,
                    PRIMARY KEY (id, topic, subscription)
                )
                """
            )
        rows = cursor.execute(
            """SELECT name FROM sqlite_master WHERE type='table'
                AND name=?""",
            (self.metadata_table,),
        ).fetchall()
        if len(rows) == 0:
            cursor.execute(
                f"""
                CREATE TABLE {self.metadata_table} (
                    topic TEXT,
                    subscription TEXT,
                    config TEXT,
                    PRIMARY KEY (topic, subscription)
                )
                """
            )
        self._client.commit()
        cursor.close()

    def _get_topic_name(self, op_parser: MessagingOperationParser) -> str:
        if self.mode == MessagingMode.PUBSUB:
            topic_name = op_parser.get_topic()
            topic_name = topic_name or self.topic or self.__component__.topic
            if topic_name is None:
                raise BadRequestError(
                    "Topic name is required for this operation.",
                )
        else:
            topic_name = op_parser.get_queue()
            topic_name = topic_name or self.queue or self.__component__.queue
            if topic_name is None:
                raise BadRequestError(
                    "Queue name is required for this operation.",
                )
        return topic_name

    def _get_subscription_name(
        self, op_parser: MessagingOperationParser
    ) -> str:
        subscription_name = op_parser.get_subscription()
        subscription_name = (
            subscription_name
            or self.subscription
            or self.__component__.subscription
        )
        if subscription_name is None:
            raise BadRequestError(
                "Subscription name is required for this operation.",
            )
        return subscription_name

    def __run__(
        self,
        operation: Operation | None = None,
        context: Context | None = None,
        **kwargs,
    ) -> Any:
        self.__setup__(context=context)
        op_parser = MessagingOperationParser(operation)
        ncall, state = self._get_ncall(
            op_parser,
            self._client_helper,
        )
        if ncall is None:
            return super().__run__(
                operation,
                context,
                **kwargs,
            )
        nresult = ncall.invoke()
        result = self._convert_nresult(
            nresult,
            state,
            op_parser,
        )
        return Response(result=result, native=dict(result=nresult, call=ncall))

    def _get_ncall(
        self,
        op_parser: MessagingOperationParser,
        helper: ClientHelper,
    ) -> tuple[NCall | None, dict]:
        call = None
        state: dict = {}
        op_converter = self._op_converter
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            args = op_converter.convert_create_queue(
                self._get_topic_name(op_parser),
                op_parser.get_queue_config(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.transact,
                args,
                None,
                {sqlite3.IntegrityError: ConflictError},
            )
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            args = op_converter.convert_update_queue(
                self._get_topic_name(op_parser),
                op_parser.get_queue_config(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            args = op_converter.convert_drop_topic(
                self._get_topic_name(op_parser),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            args = op_converter.convert_has_topic(
                self._get_topic_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            args = {
                "queue": self._get_topic_name(op_parser),
            }
            call = NCall(
                helper.get_queue,
                args,
            )
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            args = op_converter.convert_list_topics()
            call = NCall(
                helper.execute,
                args,
            )
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            args = op_converter.convert_create_topic(
                self._get_topic_name(op_parser),
                op_parser.get_topic_config(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: ConflictError},
            )
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            args = op_converter.convert_update_topic(
                self._get_topic_name(op_parser),
                op_parser.get_topic_config(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            args = op_converter.convert_drop_topic(
                self._get_topic_name(op_parser),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            args = op_converter.convert_has_topic(
                self._get_topic_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            args = {
                "topic": self._get_topic_name(op_parser),
            }
            call = NCall(
                helper.get_topic,
                args,
            )
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            args = op_converter.convert_list_topics()
            call = NCall(
                helper.execute,
                args,
            )
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            args = op_converter.convert_create_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_subscription_config(),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
                {sqlite3.IntegrityError: ConflictError},
            )
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            args = op_converter.convert_update_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_subscription_config(),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            args = op_converter.convert_drop_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_where_exists(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            args = op_converter.convert_has_subscription(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
                None,
            )
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
            }
            call = NCall(
                helper.get_subscription,
                args,
            )
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            args = op_converter.convert_list_subscriptions(
                self._get_topic_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            args = {
                "topic": self._get_topic_name(op_parser),
                "value": op_parser.get_value(),
                "metadata": op_parser.get_metadata(),
                "properties": op_parser.get_properties(),
                "config": op_parser.get_put_config(),
            }
            call = NCall(
                helper.put,
                args,
            )
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            args = {
                "topic": self._get_topic_name(op_parser),
                "batch": op_parser.get_batch(),
            }
            call = NCall(
                helper.batch,
                args,
            )
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "config": op_parser.get_pull_config(),
            }
            call = NCall(
                helper.pull,
                args,
            )
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            args = op_converter.convert_ack(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_key(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            args = op_converter.convert_nack(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
                op_parser.get_key(),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            args = {
                "topic": self._get_topic_name(op_parser),
                "subscription": self._get_subscription_name(op_parser),
                "key": op_parser.get_key(),
                "timeout": op_parser.get_timeout(),
            }
            call = NCall(
                helper.extend,
                args,
            )
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            args = op_converter.convert_purge(
                self._get_topic_name(op_parser),
                self._get_subscription_name(op_parser),
            )
            call = NCall(
                helper.execute,
                args,
            )
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            args = {
                "topic": op_parser.get_topic() or op_parser.get_queue(),
            }
            call = NCall(helper.close, args)
        return call, state

    def _convert_nresult(
        self,
        nresult: Any,
        state: dict,
        op_parser: MessagingOperationParser,
    ) -> Any:
        result: Any = None
        result_converter = self._result_converter
        # CREATE QUEUE
        if op_parser.op_equals(MessagingOperation.CREATE_QUEUE):
            result = None
        # UPDATE QUEUE
        elif op_parser.op_equals(MessagingOperation.UPDATE_QUEUE):
            if nresult == 0:
                raise NotFoundError(
                    "Queue not found.",
                )
            result = None
        # DROP QUEUE
        elif op_parser.op_equals(MessagingOperation.DROP_QUEUE):
            where_exists = op_parser.get_where_exists()
            if nresult == 0 and where_exists is True:
                raise NotFoundError
            result = None
        # HAS QUEUE
        elif op_parser.op_equals(MessagingOperation.HAS_QUEUE):
            return len(nresult) > 0
        # GET QUEUE
        elif op_parser.op_equals(MessagingOperation.GET_QUEUE):
            result = nresult
        # LIST QUEUES
        elif op_parser.op_equals(MessagingOperation.LIST_QUEUES):
            result = result_converter.convert_list_topics(nresult)
        # CREATE TOPIC
        elif op_parser.op_equals(MessagingOperation.CREATE_TOPIC):
            result = None
        # UPDATE TOPIC
        elif op_parser.op_equals(MessagingOperation.UPDATE_TOPIC):
            if nresult == 0:
                raise NotFoundError(
                    "Topic not found.",
                )
            result = None
        # DROP TOPIC
        elif op_parser.op_equals(MessagingOperation.DROP_TOPIC):
            where_exists = op_parser.get_where_exists()
            if nresult == 0 and where_exists is True:
                raise NotFoundError
            result = None
        # HAS TOPIC
        elif op_parser.op_equals(MessagingOperation.HAS_TOPIC):
            return len(nresult) > 0
        # GET TOPIC
        elif op_parser.op_equals(MessagingOperation.GET_TOPIC):
            result = nresult
        # LIST TOPICS
        elif op_parser.op_equals(MessagingOperation.LIST_TOPICS):
            result = result_converter.convert_list_topics(nresult)
        # CREATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.CREATE_SUBSCRIPTION):
            result = None
        # UPDATE SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.UPDATE_SUBSCRIPTION):
            if nresult == 0:
                raise NotFoundError(
                    "Subscription not found.",
                )
            result = None
        # DROP SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.DROP_SUBSCRIPTION):
            where_exists = op_parser.get_where_exists()
            if nresult == 0 and where_exists is True:
                raise NotFoundError
            result = None
        # HAS SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.HAS_SUBSCRIPTION):
            return len(nresult) > 0
        # GET SUBSCRIPTION
        elif op_parser.op_equals(MessagingOperation.GET_SUBSCRIPTION):
            result = nresult
        # LIST SUBSCRIPTIONS
        elif op_parser.op_equals(MessagingOperation.LIST_SUBSCRIPTIONS):
            result = result_converter.convert_list_subscriptions(nresult)
        # PUT
        elif op_parser.op_equals(MessagingOperation.PUT):
            result = None
        # BATCH
        elif op_parser.op_equals(MessagingOperation.BATCH):
            result = None
        # PULL
        elif op_parser.op_equals(MessagingOperation.PULL):
            result = nresult
        # ACK
        elif op_parser.op_equals(MessagingOperation.ACK):
            if nresult == 0:
                raise BadRequestError(
                    "Message not found or already acknowledged.",
                )
            result = None
        # NACK
        elif op_parser.op_equals(MessagingOperation.NACK):
            if nresult == 0:
                raise BadRequestError(
                    "Message not found or already acknowledged.",
                )
            result = None
        # EXTEND
        elif op_parser.op_equals(MessagingOperation.EXTEND):
            if nresult == 0:
                raise BadRequestError(
                    "Message not found or already acknowledged.",
                )
            result = MessageItem(key=op_parser.get_key())
        # PURGE
        elif op_parser.op_equals(MessagingOperation.PURGE):
            result = None
        # CLOSE
        elif op_parser.op_equals(MessagingOperation.CLOSE):
            pass
        return result


class ClientHelper:
    client: Any
    op_converter: OperationConverter
    result_converter: ResultConverter
    mode: MessagingMode
    poll_interval: float
    subscription_config_cache: dict[str, dict[str, SubscriptionConfig | None]]

    def __init__(
        self,
        client: Any,
        op_converter: OperationConverter,
        result_converter: ResultConverter,
        mode: MessagingMode,
        poll_interval: float,
        subscription_config_cache: dict[
            str, dict[str, SubscriptionConfig | None]
        ],
    ):
        self.client = client
        self.op_converter = op_converter
        self.result_converter = result_converter
        self.mode = mode
        self.poll_interval = poll_interval
        self.subscription_config_cache = subscription_config_cache

    def _get_subscription_config(
        self,
        topic_name: str,
        subscription_name: str,
    ) -> SubscriptionConfig | None:
        if (
            topic_name in self.subscription_config_cache
            and subscription_name in self.subscription_config_cache[topic_name]
        ):
            return self.subscription_config_cache[topic_name][
                subscription_name
            ]
        args = self.op_converter.convert_get_subscription_config(
            topic_name,
            subscription_name,
        )
        nresult = self.execute(
            args["query"],
            args["params"],
            fetchall=True,
        )
        if nresult is None or len(nresult) == 0:
            raise BadRequestError(
                f"Subscription {subscription_name} not found.",
            )
        if nresult[0][0] is None:
            config = None
        else:
            config = SubscriptionConfig.from_dict(json.loads(nresult[0][0]))
        if topic_name not in self.subscription_config_cache:
            self.subscription_config_cache[topic_name] = {}
        self.subscription_config_cache[topic_name][subscription_name] = config
        return config

    def get_queue(self, queue: str) -> QueueInfo:
        sinfo = self.get_subscription(queue, DEFAULT_SUBSCRIPTION_NAME)
        return QueueInfo(
            name=queue,
            active_message_count=sinfo.active_message_count,
            inflight_message_count=sinfo.inflight_message_count,
            scheduled_message_count=sinfo.scheduled_message_count,
            config=(
                QueueConfig.from_dict(sinfo.config.to_dict())
                if sinfo.config
                else None
            ),
            nref=queue,
        )

    def get_topic(self, topic: str) -> TopicInfo:
        args = self.op_converter.convert_get_topic_config(topic)
        nresult = self.execute(args["query"], args["params"])
        if nresult is None:
            raise NotFoundError
        config_json = nresult[0]
        args = self.op_converter.convert_get_subscription_count(topic)
        subscription_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        return TopicInfo(
            name=topic,
            subscription_count=subscription_count - 1,
            config=(
                TopicConfig.from_dict(json.loads(config_json))
                if config_json
                else None
            ),
            nref=topic,
        )

    def get_subscription(
        self, topic: str, subscription: str
    ) -> SubscriptionInfo:
        args = self.op_converter.convert_get_subscription_config(
            topic, subscription
        )
        nresult = self.execute(args["query"], args["params"])
        if nresult is None:
            raise NotFoundError
        config_json = nresult[0]
        args = self.op_converter.convert_get_active_message_count(
            topic, subscription
        )
        active_message_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        args = self.op_converter.convert_get_inflight_message_count(
            topic, subscription
        )
        inflight_message_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        args = self.op_converter.convert_get_scheduled_message_count(
            topic, subscription
        )
        scheduled_message_count = self.execute(
            args["query"], args["params"], fetchall=True
        )[0][0]
        return SubscriptionInfo(
            name=subscription,
            topic=topic,
            active_message_count=active_message_count,
            inflight_message_count=inflight_message_count,
            scheduled_message_count=scheduled_message_count,
            config=(
                SubscriptionConfig.from_dict(json.loads(config_json))
                if config_json
                else None
            ),
            nref=subscription,
        )

    def put(
        self,
        topic: str,
        value: MessageValueType,
        metadata: dict | None = None,
        properties: MessageProperties | None = None,
        config: MessagePutConfig | None = None,
    ) -> None:
        args = self.op_converter.convert_list_subscriptions(topic)
        nresult = self.client.execute(args["query"], args["params"]).fetchall()
        subscriptions = self.result_converter.convert_list_subscriptions(
            nresult
        )
        if len(subscriptions) == 0:
            if self.mode == MessagingMode.QUEUE:
                raise BadRequestError(
                    f"Queue {topic} not found.",
                )
        if properties and properties.message_id:
            id = properties.message_id
        else:
            id = str(uuid.uuid4())
        enqueue_time = Time.now()
        if config and config.delay:
            enqueue_time += config.delay
        ops: list = []
        for subscription in subscriptions:
            args = self.op_converter.convert_put(
                id,
                topic,
                subscription,
                enqueue_time,
                value,
                metadata,
                properties,
            )
            ops.append(args)
        self.transact(ops=ops)

    def batch(
        self,
        topic: str,
        batch: MessageBatch,
    ) -> None:
        for operation in batch.operations:
            op_parser = MessagingOperationParser(operation)
            self.put(
                topic=topic,
                value=op_parser.get_value(),
                metadata=op_parser.get_metadata(),
                properties=op_parser.get_properties(),
                config=op_parser.get_put_config(),
            )

    def pull(
        self,
        topic: str,
        subscription: str,
        config: MessagePullConfig | None = None,
    ) -> list[MessageItem]:
        subscription_config = self._get_subscription_config(
            topic,
            subscription,
        )
        result: list[MessageItem] = []
        start_time = Time.now()
        while True:
            now = Time.now()
            found = False
            with self.client:
                args = self.op_converter.convert_pull(
                    topic,
                    subscription,
                    subscription_config=subscription_config,
                )
                nresult = self.client.execute(
                    args["query"], args["params"]
                ).fetchone()
                if nresult is not None:
                    lock_token = str(uuid.uuid4())
                    message = self.result_converter.convert_pull(nresult)
                    if message.key is None or message.key.id is None:
                        raise BadRequestError(
                            "Message key is missing.",
                        )
                    lock_duration = (
                        config.visibility_timeout if config else None
                    )
                    lock_duration = lock_duration or (
                        subscription_config.visibility_timeout
                        if subscription_config
                        and subscription_config.visibility_timeout
                        else DEFAULT_VISIBILITY_TIMEOUT
                    )
                    lock_until_time = now + lock_duration
                    args = self.op_converter.convert_lock(
                        message.key.id,
                        topic,
                        subscription,
                        lock_until_time,
                        lock_token,
                        message.key.nref,
                    )
                    updated = self.client.execute(
                        args["query"], args["params"]
                    ).rowcount
                    if updated == 1:
                        message.key.nref = lock_token
                        result.append(message)
                        found = True
            if config and config.max_wait_time:
                if now - start_time > config.max_wait_time:
                    break
            if config and config.max_count:
                if len(result) >= config.max_count:
                    break
            if not config or not config.max_count:
                if len(result) > 0:
                    break
            if not found:
                time.sleep(self.poll_interval)
        return result

    def extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None = None,
    ) -> None:
        subscription_config = self._get_subscription_config(
            topic, subscription
        )
        args = self.op_converter.convert_extend(
            topic,
            subscription,
            key,
            timeout,
            subscription_config,
        )
        self.client.execute(args["query"], args["params"])

    def execute(
        self,
        query: str,
        params: tuple | None = None,
        rowcount: bool = False,
        fetchall: bool = False,
    ) -> Any:
        cursor = self.client.cursor()
        try:
            if params:
                cursor.execute(query, params)
            else:
                cursor.execute(query)
            if rowcount:
                return cursor.rowcount
            if fetchall:
                return cursor.fetchall()
            return cursor.fetchone()
        finally:
            self.client.commit()
            cursor.close()

    def transact(self, ops: list) -> Any:
        result = []
        cursor = self.client.cursor()
        try:
            for op in ops:
                if "params" in op:
                    cursor.execute(op["query"], op["params"])
                else:
                    cursor.execute(op["query"])
                if "rowcount" in op and op["rowcount"]:
                    if cursor.rowcount == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(cursor.rowcount)
                else:
                    res = cursor.fetchone()
                    if res is None or len(res) == 0:
                        self.client.rollback()
                        raise ConflictError
                    result.append(res)
            self.client.commit()
        finally:
            cursor.close()
        return result

    def close(self, topic: str | None) -> Any:
        if topic is not None:
            return
        self.client.close()


class OperationConverter:
    message_table: str
    metadata_table: str

    META_SUBSCRIPTION_NAME = "#"

    def __init__(
        self,
        message_table: str,
        metadata_table: str,
    ):
        self.message_table = message_table
        self.metadata_table = metadata_table

    def convert_get_active_message_count(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        current_time = Time.now()
        query = f"""
            SELECT COUNT(*)
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_until_time < ?
            AND enqueued_time <= ?
            """
        params = (
            topic,
            subscription,
            current_time,
            current_time,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_inflight_message_count(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        current_time = Time.now()
        query = f"""
            SELECT COUNT(*)
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_until_time >= ?
            """
        params = (
            topic,
            subscription,
            current_time,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_scheduled_message_count(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        current_time = Time.now()
        query = f"""
            SELECT COUNT(*)
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND enqueued_time > ?
            """
        params = (
            topic,
            subscription,
            current_time,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_subscription_count(
        self,
        topic: str,
    ) -> dict:
        query = f"""
            SELECT COUNT(*)
            FROM {self.metadata_table}
            WHERE topic = ?
            """
        params = (topic,)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_get_topic_config(
        self,
        topic: str,
    ) -> dict:
        query = f"""
            SELECT config
            FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, OperationConverter.META_SUBSCRIPTION_NAME)
        return {
            "query": query,
            "params": params,
        }

    def convert_get_subscription_config(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        query = f"""
            SELECT config
            FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
        }

    def convert_create_queue(
        self,
        queue: str,
        config: QueueConfig | None,
        where_exists: bool | None,
    ) -> dict:
        ops: list = []
        args = self.convert_create_topic(
            queue,
            None,
            where_exists,
        )
        ops.append(args)
        args = self.convert_create_subscription(
            queue,
            DEFAULT_SUBSCRIPTION_NAME,
            config,
            where_exists,
        )
        ops.append(args)
        return {"ops": ops}

    def convert_update_queue(
        self,
        queue: str,
        config: QueueConfig | None,
    ) -> dict:
        return self.convert_update_subscription(
            queue,
            DEFAULT_SUBSCRIPTION_NAME,
            config,
        )

    def convert_create_topic(
        self,
        topic: str,
        config: QueueConfig | TopicConfig | None,
        where_exists: bool | None,
    ) -> dict:
        if where_exists is False:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                """
        else:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                ON CONFLICT (topic, subscription)
                DO UPDATE SET
                config = EXCLUDED.config
                """
        params = (
            topic,
            OperationConverter.META_SUBSCRIPTION_NAME,
            config.to_json() if config else None,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_update_topic(
        self,
        topic: str,
        config: TopicConfig | None,
    ) -> dict:
        query = f"""
            UPDATE {self.metadata_table}
            SET config = ?
            WHERE topic = ? AND subscription = ?
            """
        params = (
            config.to_json() if config else None,
            topic,
            OperationConverter.META_SUBSCRIPTION_NAME,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_drop_topic(
        self,
        topic: str,
        where_exists: bool | None,
    ) -> dict:
        query = f"""
            DELETE FROM {self.metadata_table}
            WHERE topic = ?
            """
        params = (topic,)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_has_topic(
        self,
        topic: str,
    ) -> dict:
        query = f"""
            SELECT topic
            FROM {self.metadata_table}
            WHERE topic = ?
            """
        params = (topic,)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_list_topics(self) -> dict:
        query = f"""
            SELECT topic
            FROM {self.metadata_table}
            WHERE subscription = ?
            """
        params = (OperationConverter.META_SUBSCRIPTION_NAME,)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_create_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | QueueConfig | None,
        where_exists: bool | None,
    ) -> dict:
        if where_exists is False:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                """
        else:
            query = f"""
                INSERT INTO {self.metadata_table}
                (topic, subscription, config)
                VALUES (?, ?, ?)
                ON CONFLICT (topic, subscription)
                DO UPDATE SET
                config = EXCLUDED.config
                """
        params = (
            topic,
            subscription,
            config.to_json() if config else None,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_update_subscription(
        self,
        topic: str,
        subscription: str,
        config: SubscriptionConfig | QueueConfig | None,
    ) -> dict:
        query = f"""
            UPDATE {self.metadata_table}
            SET config = ?
            WHERE topic = ? AND subscription = ?
            """
        params = (
            config.to_json() if config else None,
            topic,
            subscription,
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_drop_subscription(
        self,
        topic: str,
        subscription: str,
        where_exists: bool | None,
    ) -> dict:
        query = f"""
            DELETE FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_has_subscription(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        query = f"""
            SELECT topic, subscription
            FROM {self.metadata_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_list_subscriptions(self, topic: str) -> dict:
        query = f"""
            SELECT subscription
            FROM {self.metadata_table}
            WHERE topic = ?
            AND subscription != ?
            """
        params = (
            topic,
            OperationConverter.META_SUBSCRIPTION_NAME,
        )
        return {
            "query": query,
            "params": params,
            "fetchall": True,
        }

    def convert_put(
        self,
        id: str,
        topic: str,
        subscription: str,
        enqueue_time: float,
        value: MessageValueType,
        metadata: dict | None,
        properties: MessageProperties | None,
    ) -> dict:
        message_id = None
        content_type = None
        group_id = None
        body: bytes | None = None
        if properties is not None:
            message_id = properties.message_id
            content_type = properties.content_type
            group_id = properties.group_id
        if isinstance(value, str):
            body = value.encode("utf-8")
            content_type = "text/plain"
        elif isinstance(value, bytes):
            body = value
            content_type = "application/octet-stream"
        elif isinstance(value, dict):
            body = json.dumps(value).encode("utf-8")
            content_type = "application/json"
        elif isinstance(value, DataModel):
            body = value.to_json().encode("utf-8")
            content_type = "application/json"
        else:
            raise BadRequestError("Message type not supported")
        query = f"""
            INSERT INTO {self.message_table}
            (id, topic, subscription, value, metadata,
            message_id, group_id, content_type,
            enqueued_time, delivery_count,
            lock_until_time, lock_token)
            VALUES (?, ?, ?, ?, ?,
            ?, ?, ?, ?, ?, ?, ?)
            """
        params = (
            id,
            topic,
            subscription,
            body,
            json.dumps(metadata) if metadata else None,
            message_id,
            group_id,
            content_type,
            enqueue_time,
            0,
            0,
            "",
        )
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_pull(
        self,
        topic: str,
        subscription: str,
        subscription_config: SubscriptionConfig | None = None,
    ) -> dict:
        ttl = (
            subscription_config.ttl
            if subscription_config and subscription_config.ttl
            else DEFAULT_TTL
        )
        current_time = Time.now()
        query = f"""
            SELECT id, topic, subscription, value, metadata,
            message_id, group_id, content_type,
            enqueued_time, delivery_count,
            lock_until_time, lock_token
            FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_until_time <= ?
            AND enqueued_time <= ?
            AND enqueued_time >= ?
            ORDER BY enqueued_time
            LIMIT 1
            """
        params = (
            topic,
            subscription,
            current_time,
            current_time,
            current_time - ttl,
        )
        return {
            "query": query,
            "params": params,
        }

    def convert_lock(
        self,
        id: str,
        topic: str,
        subscription: str,
        lock_until_time: float,
        lock_token: str,
        previous_lock_token: str | None,
    ) -> dict:
        query = f"""
            UPDATE {self.message_table}
            SET lock_until_time = ?, lock_token = ?,
            delivery_count = delivery_count + 1
            WHERE topic = ? AND subscription = ?
            AND id = ? AND lock_token = ?
            """
        params = (
            lock_until_time,
            lock_token,
            topic,
            subscription,
            id,
            previous_lock_token,
        )
        return {
            "query": query,
            "params": params,
        }

    def convert_ack(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
    ) -> dict:
        query = f"""
            DELETE FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            AND lock_token = ?
            """
        params = (topic, subscription, key.nref)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_nack(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
    ) -> dict:
        query = f"""
            UPDATE {self.message_table}
            SET lock_until_time = ?, lock_token = ?
            WHERE topic = ? AND subscription = ?
            AND lock_token = ?
            """
        params = (0, "", topic, subscription, key.nref)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_extend(
        self,
        topic: str,
        subscription: str,
        key: MessageKey,
        timeout: int | None,
        subscription_config: SubscriptionConfig | None = None,
    ) -> dict:
        lock_duration = (
            subscription_config.visibility_timeout
            if subscription_config and subscription_config.visibility_timeout
            else DEFAULT_VISIBILITY_TIMEOUT
        )
        t = timeout or lock_duration
        t = Time.now() + t
        query = f"""
            UPDATE {self.message_table}
            SET lock_until_time = ?
            WHERE topic = ? AND subscription = ?
            AND lock_token = ?
            """
        params = (t, topic, subscription, key.nref)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }

    def convert_purge(
        self,
        topic: str,
        subscription: str,
    ) -> dict:
        query = f"""
            DELETE FROM {self.message_table}
            WHERE topic = ? AND subscription = ?
            """
        params = (topic, subscription)
        return {
            "query": query,
            "params": params,
            "rowcount": True,
        }


class ResultConverter:
    def __init__(self):
        pass

    def _convert_value(
        self,
        value: bytes,
        content_type: str | None,
    ) -> MessageValueType:
        if content_type == "text/plain":
            return value.decode("utf-8")
        elif content_type == "application/octet-stream":
            return value
        elif content_type == "application/json":
            return json.loads(value.decode("utf-8"))
        return value

    def convert_list_topics(self, nresult: Any) -> list[str]:
        topics: list[str] = []
        for row in nresult:
            topics.append(row[0])
        return topics

    def convert_list_subscriptions(self, nresult: Any) -> list[str]:
        subscriptions: list[str] = []
        for row in nresult:
            subscriptions.append(row[0])
        return subscriptions

    def convert_pull(
        self,
        nresult: Any,
    ) -> MessageItem:
        id = nresult[0]
        value = nresult[3]
        metadata = nresult[4]
        message_id = nresult[5]
        group_id = nresult[6]
        content_type = nresult[7]
        enqueued_time = nresult[8]
        delivery_count = nresult[9]
        lock_token = nresult[11]

        message = MessageItem(
            key=MessageKey(id=id, nref=lock_token),
            value=self._convert_value(
                value,
                content_type,
            ),
            metadata=json.loads(metadata) if metadata else None,
            properties=MessageProperties(
                message_id=message_id,
                group_id=group_id,
                content_type=content_type,
                enqueued_time=enqueued_time,
                delivery_count=delivery_count + 1,
            ),
        )
        return message



import time

from x8.content.video import Video

from ._providers import get_component


def run():
    client = get_component(provider_type="openai")
    res = client.generate(prompt="Hamsters doing cartwheels on snow in Mars")
    print(res)

    while True:
        res = client.get(res.result.id)
        print("-------")
        print(res)
        if res.result.status in ["completed", "failed"]:
            break
        time.sleep(1)

    res = client.download(res.result.id)
    video = Video.load(res.result)
    video.save("a.mp4")
    print(video)


if __name__ == "__main__":
    run()



from x8.core import Component


class Web(Component):
    component: Component | None

    def __init__(
        self,
        component: Component | None = None,
        **kwargs,
    ):
        """Initialize."""
        self.component = component
        super().__init__(**kwargs)
